#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Login Command Module."""

from __future__ import annotations

import os
import shlex
import subprocess

from typing import TYPE_CHECKING

from molecule import scenarios
from molecule.click_cfg import click_command_ex, options
from molecule.command import base
from molecule.exceptions import MoleculeError


if TYPE_CHECKING:
    import click

    from molecule import config
    from molecule.types import CommandArgs


class Login(base.Base):
    """Login Command Class."""

    def __init__(self, c: config.Config) -> None:
        """Construct Login.

        Args:
            c: An instance of a Molecule config.
        """
        super().__init__(c)
        self._pt = None

    def execute(self, action_args: list[str] | None = None) -> None:  # noqa: ARG002
        """Execute the actions necessary to perform a `molecule login`.

        Args:
            action_args: Arguments for this command. Unused.
        """
        c = self._config
        if (not c.state.created) and c.driver.managed:
            base.execute_subcommand(c, "create")

        hosts = [d["name"] for d in self._config.platforms.instances]
        hostname = self._get_hostname(hosts)
        self._get_login(hostname)

    def _get_hostname(self, hosts: list[str]) -> str:
        hostname = self._config.command_args.get("host")
        host_list = "\n".join(sorted(hosts))
        if hostname is None:
            if len(hosts) == 1:
                hostname = hosts[0]
            else:
                msg = (
                    f"There are {len(hosts)} running hosts. Please specify "
                    "which with --host.\n\n"
                    f"Available hosts:\n{host_list}"
                )
                raise MoleculeError(msg)
        match = [x for x in hosts if x.startswith(hostname)]
        if len(match) == 0:
            msg = f"There are no hosts that match '{hostname}'.  You can only login to valid hosts."
            raise MoleculeError(msg)
        if len(match) != 1:
            # If there are multiple matches, but one of them is an exact string
            # match, assume this is the one they're looking for and use it.
            if hostname in match:
                match = [hostname]
            else:
                msg = (
                    f"There are {len(match)} hosts that match '{hostname}'. You "
                    "can only login to one at a time.\n\n"
                    f"Available hosts:\n{host_list}"
                )
                raise MoleculeError(msg)

        return match[0]

    def _get_login(self, hostname: str) -> None:  # pragma: no cover
        # ruff: noqa: S605
        lines, columns = os.popen("stty size", "r").read().split()
        login_options = self._config.driver.login_options(hostname)
        login_options["columns"] = columns
        login_options["lines"] = lines
        if not self._config.driver.login_cmd_template:
            self._log.warning(
                "Login command is not supported for [dim]%s[/] host because 'login_cmd_template' was not defined in driver options.",
                login_options["instance"],
            )
            return
        login_cmd = self._config.driver.login_cmd_template.format(**login_options)

        cmd = shlex.split(f"/usr/bin/env {login_cmd}")
        subprocess.run(cmd, check=False)


@click_command_ex()
@options(["host", "scenario_name_single_with_default"])
def login(ctx: click.Context) -> None:  # pragma: no cover
    """Log in to one instance.

    Args:
        ctx: Click context object holding commandline arguments.
    """
    args = ctx.obj.get("args")
    subcommand = base._get_subcommand(__name__)  # noqa: SLF001
    command_args: CommandArgs = {
        "host": ctx.params["host"],
        "subcommand": subcommand,
    }

    scenario_name = ctx.params["scenario_name"]

    s = scenarios.Scenarios(base.get_configs(args, command_args), [scenario_name])
    for scenario in s.all:
        base.execute_subcommand(scenario.config, subcommand)



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

from subprocess import CompletedProcess

import pytest

from molecule import config
from molecule.exceptions import MoleculeError
from molecule.provisioner import ansible_playbook


@pytest.fixture
def _instance(config_instance: config.Config) -> ansible_playbook.AnsiblePlaybook:
    _instance = ansible_playbook.AnsiblePlaybook("playbook", config_instance)

    return _instance  # noqa: RET504


@pytest.fixture
def _provisioner_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"provisioner": {"name": "ansible", "env": {"FOO": "bar"}}}


@pytest.fixture
def _verifier_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"verifier": {"name": "ansible", "env": {"FOO": "bar"}}}


@pytest.fixture
def _provisioner_verifier_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {
        "provisioner": {"name": "ansible", "env": {"FOO": "bar"}},
        "verifier": {"name": "ansible", "env": {"FOO": "baz"}},
    }


@pytest.fixture
def _instance_for_verifier_env(config_instance: config.Config):  # type: ignore[no-untyped-def]  # noqa: ANN202
    _instance = ansible_playbook.AnsiblePlaybook(
        "playbook",
        config_instance,
        verify=True,
    )
    return _instance  # noqa: RET504


@pytest.mark.parametrize(
    "config_instance",
    ["_provisioner_section_data"],  # noqa: PT007
    indirect=True,
)
def test_env_in_provision(_instance_for_verifier_env):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance_for_verifier_env._env["FOO"] == "bar"


@pytest.mark.parametrize(
    "config_instance",
    ["_verifier_section_data"],  # noqa: PT007
    indirect=True,
)
def test_env_in_verifier(_instance_for_verifier_env):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance_for_verifier_env._env["FOO"] == "bar"


@pytest.mark.parametrize(
    "config_instance",
    ["_provisioner_verifier_section_data"],  # noqa: PT007
    indirect=True,
)
def test_env_in_verify_override_provision(_instance_for_verifier_env):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance_for_verifier_env._env["FOO"] == "baz"


@pytest.fixture
def _inventory_directory(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN202
    return _instance._config.provisioner.inventory_directory


def test_ansible_command_private_member(  # noqa: D103
    _instance: ansible_playbook.AnsiblePlaybook,  # noqa: PT019
) -> None:
    assert _instance._ansible_command == []


def test_ansible_playbook_private_member(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance._playbook == "playbook"


def test_config_private_member(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert isinstance(_instance._config, config.Config)


def test_bake(_inventory_directory, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    pb = _instance._config.provisioner.playbooks.converge
    _instance._playbook = pb
    _instance.bake()

    args = [
        "ansible-playbook",
        "--become",
        "--inventory",
        _inventory_directory,
        "--skip-tags",
        "molecule-notest,notest",
        pb,
    ]

    assert _instance._ansible_command == args


def test_bake_with_ansible_navigator(_inventory_directory, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    pb = _instance._config.provisioner.playbooks.converge
    _instance._playbook = pb
    _instance._config.config["executor"]["backend"] = "ansible-navigator"
    _instance.bake()

    args = [
        "ansible-navigator",
        "run",
        pb,
        "--mode",
        "stdout",
        "--become",
        "--inventory",
        _inventory_directory,
        "--skip-tags",
        "molecule-notest,notest",
    ]

    assert _instance._ansible_command == args


def test_bake_removes_non_interactive_options_from_non_converge_playbooks(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    _inventory_directory,  # noqa: PT019
    _instance,  # noqa: PT019
):
    _instance.bake()

    args = [
        "ansible-playbook",
        "--inventory",
        _inventory_directory,
        "--skip-tags",
        "molecule-notest,notest",
        "playbook",
    ]

    assert _instance._ansible_command == args


def test_bake_has_ansible_args(_inventory_directory, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    _instance._config.ansible_args = ("foo", "bar")
    _instance._config.config["provisioner"]["ansible_args"] = ("frob", "nitz")
    _instance.bake()

    args = [
        "ansible-playbook",
        "--inventory",
        _inventory_directory,
        "--skip-tags",
        "molecule-notest,notest",
        "frob",
        "nitz",
        "foo",
        "bar",
        "playbook",
    ]

    assert _instance._ansible_command == args


def test_bake_does_not_have_ansible_args(_inventory_directory, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    for action in ["create", "destroy"]:
        _instance._config.ansible_args = ("foo", "bar")
        _instance._config.action = action
        _instance.bake()

        args = [
            "ansible-playbook",
            "--inventory",
            _inventory_directory,
            "--skip-tags",
            "molecule-notest,notest",
            "playbook",
        ]

        assert _instance._ansible_command == args


def test_bake_idem_does_have_skip_tag(_inventory_directory, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    _instance._config.action = "idempotence"
    _instance.bake()

    args = [
        "ansible-playbook",
        "--inventory",
        _inventory_directory,
        "--skip-tags",
        "molecule-notest,notest,molecule-idempotence-notest",
        "playbook",
    ]

    assert _instance._ansible_command == args


def test_execute_playbook(patched_run_command, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    _instance._ansible_command = "patched-command"
    result = _instance.execute()
    assert result == "patched-run-command-stdout"


def test_ansible_execute_bakes(_inventory_directory, patched_run_command, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    _instance.execute()

    assert _instance._ansible_command is not None

    args = [
        "ansible-playbook",
        "--inventory",
        _inventory_directory,
        "--skip-tags",
        "molecule-notest,notest",
        "playbook",
    ]

    assert _instance._ansible_command == args


def test_execute_bakes_with_ansible_args(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    _inventory_directory,  # noqa: PT019
    patched_run_command,
    _instance,  # noqa: PT019
):
    _instance._config.ansible_args = ("-o", "--syntax-check")
    _instance.execute()

    assert _instance._ansible_command is not None

    args = [
        "ansible-playbook",
        "--inventory",
        _inventory_directory,
        "--skip-tags",
        "molecule-notest,notest",
        "-o",
        "--syntax-check",
        "playbook",
    ]

    assert _instance._ansible_command == args


def test_executes_catches_and_exits_return_code(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    patched_run_command,
    _instance,  # noqa: PT019
):
    patched_run_command.side_effect = [
        CompletedProcess(
            args="ansible-playbook",
            returncode=1,
            stdout="out",
            stderr="err",
        ),
    ]
    with pytest.raises(MoleculeError) as e:
        _instance.execute()

    assert e.value.code == 1


def test_add_cli_arg(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance._cli == {}

    _instance.add_cli_arg("foo", "bar")
    assert _instance._cli == {"foo": "bar"}


def test_add_env_arg(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert "foo" not in _instance._env

    _instance.add_env_arg("foo", "bar")
    assert _instance._env["foo"] == "bar"



"""ANSI Output formatting and color support for Molecule.

This module provides a reusable class for converting Rich-style markup
to ANSI escape codes while respecting color configuration.

Molecule Output Color Mapping

This color scheme is designed to align closely with Ansible's own terminal color conventions,
providing consistent visual language across Molecule and Ansible output. Key colors such as
GREEN for success (OK), RED for errors, YELLOW for changes, and MAGENTA for warnings reflect
Ansible's defaults as defined in `ansible/config/base.yml`. Other colors—such as CYAN and BLUE—
are used for clarity, neutrality, or emphasis in Molecule-specific output like actions,
scenario names, and commands.

This approach ensures readability, terminal compatibility, and a familiar experience for users
already accustomed to Ansible's interface.
"""

from __future__ import annotations

import os
import re
import sys


def to_bool(a: object) -> bool:
    """Return a bool for the arg.

    Args:
        a: A value to coerce to bool.

    Returns:
        A bool representation of a.
    """
    if a is None or isinstance(a, bool):
        return bool(a)
    if isinstance(a, str):
        a = a.lower()
    return a in ("yes", "on", "1", "true", 1)


def should_do_markup() -> bool:
    """Decide about use of ANSI colors.

    Returns:
        Whether the output should be colored.
    """
    py_colors = None

    # https://xkcd.com/927/
    for v in ["PY_COLORS", "CLICOLOR", "FORCE_COLOR", "ANSIBLE_FORCE_COLOR"]:
        value = os.environ.get(v, None)
        if value is not None:
            py_colors = to_bool(value)
            break

    # If deliberately disabled colors
    if os.environ.get("NO_COLOR", None):
        return False

    # User configuration requested colors
    if py_colors is not None:
        return to_bool(py_colors)

    term = os.environ.get("TERM", "")
    if "xterm" in term:
        return True

    if term == "dumb":
        return False

    # Use tty detection logic as last resort
    return sys.stdout.isatty()


class AnsiOutput:
    """ANSI output formatter with Rich-style markup support.

    Converts Rich-style markup tags to ANSI escape codes while respecting
    color configuration and providing reusable formatting functionality.
    """

    # ANSI Color Constants
    RESET = "\033[0m"

    # Basic Colors
    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"

    # Bright Colors
    BRIGHT_BLACK = "\033[90m"
    BRIGHT_RED = "\033[91m"
    BRIGHT_GREEN = "\033[92m"
    BRIGHT_YELLOW = "\033[93m"
    BRIGHT_BLUE = "\033[94m"
    BRIGHT_MAGENTA = "\033[95m"
    BRIGHT_CYAN = "\033[96m"
    BRIGHT_WHITE = "\033[97m"

    # Text Styles
    BOLD = "\033[1m"
    DIM = "\033[2m"
    ITALIC = "\033[3m"
    UNDERLINE = "\033[4m"

    # Symbols
    RIGHT_ARROW = "\u279c"

    def __init__(self) -> None:
        """Initialize the ANSI output formatter."""
        self.markup_enabled = should_do_markup()

        # Rich-style markup to ANSI mapping (Ansible-aligned)
        self.markup_map: dict[str, str] = {
            # ─── Message Types ───────────────────────────────
            "info": self.CYAN + self.DIM,  # Matches Ansible skip
            "warning": self.MAGENTA + self.BOLD,  # Matches Ansible warn
            "danger": self.RED + self.BOLD,  # Matches Ansible error
            "scenario": self.GREEN,  # Matches Ansible OK
            "action": self.YELLOW,  # Matches Ansible changed
            "exec_phase": self.BRIGHT_CYAN,  # Execution phase indicator
            "command": self.WHITE,
            "section_title": self.CYAN + self.BOLD,  # Neutral and distinct
            # ─── Logging Levels ──────────────────────────────
            "logging.level.debug": self.DIM + self.BLUE,
            "logging.level.info": self.CYAN + self.DIM,
            "logging.level.success": self.GREEN + self.BOLD,
            "logging.level.warning": self.MAGENTA + self.BOLD,
            "logging.level.error": self.RED + self.BOLD,
            "logging.level.critical": self.RED + self.BOLD + self.UNDERLINE,
            # Basic color names
            "red": self.RED,
            "green": self.GREEN,
            "blue": self.BLUE,
            "yellow": self.YELLOW,
            "magenta": self.MAGENTA,
            "cyan": self.CYAN,
            "white": self.WHITE,
            "black": self.BLACK,
            # Bright colors
            "bright_red": self.BRIGHT_RED,
            "bright_green": self.BRIGHT_GREEN,
            "bright_blue": self.BRIGHT_BLUE,
            "bright_yellow": self.BRIGHT_YELLOW,
            "bright_magenta": self.BRIGHT_MAGENTA,
            "bright_cyan": self.BRIGHT_CYAN,
            "bright_white": self.BRIGHT_WHITE,
            # Text styles
            "bold": self.BOLD,
            "dim": self.DIM,
            "italic": self.ITALIC,
            "underline": self.UNDERLINE,
        }

    def strip_markup(self, text: str) -> str:
        """Strip all Rich-style markup tags from text.

        Args:
            text: Text containing Rich markup tags.

        Returns:
            Text with all markup tags removed.
        """
        return re.sub(r"\[/?[^\]]*\]", "", text)

    def process_markup(self, text: str) -> str:
        """Convert Rich-style markup tags to ANSI escape codes.

        Args:
            text: Text containing Rich markup tags like [red], [bold], [/].

        Returns:
            Text with markup converted to ANSI codes or stripped if disabled.
        """
        if not self.markup_enabled:
            return self.strip_markup(text)

        # Process opening tags first to avoid conflicts with ANSI escape sequences
        def replace_tag(match: re.Match[str]) -> str:
            tag = match.group(1).lower()
            return self.markup_map.get(tag, "")

        # Match tags that don't start with /
        processed = re.sub(r"\[([^/\]]+)\]", replace_tag, text)

        # Process closing tags last (convert [/] to reset)
        result = re.sub(r"\[/\]", self.RESET, processed)
        return str(result)

    def format_log_level(self, level_name: str) -> tuple[str, str]:
        """Format a log level returning both colored and plain versions.

        Args:
            level_name: Name of the log level (e.g., 'INFO', 'WARNING').

        Returns:
            Tuple of (colored_version, plain_version).
        """
        width = 8

        if not self.markup_enabled:
            plain = f"{level_name:<{width}}"
            return plain, plain

        markup_key = f"logging.level.{level_name.lower()}"
        markup_text = f"[{markup_key}]{level_name:<{width}}[/]"
        colored = self.process_markup(markup_text)
        plain = self.strip_markup(markup_text)
        return colored, plain

    def format_scenario(self, scenario_name: str, step: str | None = None) -> tuple[str, str]:
        """Format a scenario name returning both colored and plain versions.

        Args:
            scenario_name: Name of the scenario.
            step: Optional step name (e.g., 'converge', 'create', 'destroy').

        Returns:
            Tuple of (colored_version, plain_version).
        """
        if not self.markup_enabled:
            plain = f"[{scenario_name} > {step}]" if step else f"[{scenario_name}]"
            return plain, plain

        if step:
            markup_text = rf"[scenario]{scenario_name}[/] {self.RIGHT_ARROW} [action]{step}[/]:"
        else:
            markup_text = rf"[scenario]{scenario_name}[/]"

        colored = self.process_markup(markup_text)
        plain = self.strip_markup(markup_text)
        return colored, plain



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Shell Dependency Module."""

from __future__ import annotations

from typing import TYPE_CHECKING

from molecule.dependency import base


if TYPE_CHECKING:
    from collections.abc import MutableMapping

    from molecule.config import Config


class Shell(base.Base):
    """``Shell`` is an alternate dependency manager.

    It is intended to run a command in situations where `Ansible Galaxy`_
    don't suffice.

    The ``command`` to execute is required, and is relative to Molecule's
    project directory when referencing a script not in $PATH.

    !!! note

        Unlike the other dependency managers, ``options`` are ignored and not
        passed to `shell`.  Additional flags/subcommands should simply be added
        to the `command`.

    ``` yaml
        dependency:
          name: shell
          command: path/to/command --flag1 subcommand --flag2
    ```

    The dependency manager can be disabled by setting ``enabled`` to False.

    ``` yaml
        dependency:
          name: shell
          command: path/to/command --flag1 subcommand --flag2
          enabled: False
    ```
    Environment variables can be passed to the dependency.

    ``` yaml
        dependency:
          name: shell
          command: path/to/command --flag1 subcommand --flag2
          env:
            FOO: bar
    ```
    """

    def __init__(self, config: Config) -> None:
        """Construct Shell.

        Args:
            config: Molecule Config instance.
        """
        super().__init__(config)
        self._sh_command = ""

    @property
    def command(self) -> str:
        """Return shell command.

        Returns:
            Command defined in Molecule config.
        """
        return self._config.config["dependency"]["command"] or ""

    @property
    def default_options(self) -> MutableMapping[str, str | bool]:
        """Get default options for shell dependencies (none).

        Returns:
            An empty dictionary.
        """
        return {}

    def bake(self) -> None:
        """Bake a ``shell`` command so it's ready to execute."""
        self._sh_command = self.command

    def execute(self, action_args: list[str] | None = None) -> None:  # noqa: ARG002
        """Execute the dependency solver.

        Args:
            action_args: Arguments for the dependency. Unused.
        """
        if not self.enabled:
            msg = "Skipping, dependency is disabled."
            self._log.warning(msg)
            return
        super().execute()

        if not self._sh_command:
            self.bake()
        self.execute_with_retries()

    def _has_command_configured(self) -> bool:
        return "command" in self._config.config["dependency"]



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""PLatforms Module."""

from __future__ import annotations

from typing import TYPE_CHECKING

from molecule import util


if TYPE_CHECKING:
    from molecule.config import Config
    from molecule.types import PlatformData


class Platforms:
    """Platforms define the instances to be tested, and the groups to which the instances belong.

    ``` yaml

        platforms:
          - name: instance-1
    ```

    Multiple instances can be provided.

    ``` yaml
        platforms:
          - name: instance-1
          - name: instance-2
    ```

    Mapping instances to groups.  These groups will be used by the Provisioner_
    for orchestration purposes.

    ``` yaml
        platforms:
          - name: instance-1
            groups:
              - group1
              - group2
    ```

    Children allow the creation of groups of groups.

    ``` yaml
        platforms:
          - name: instance-1
            groups:
              - group1
              - group2
            children:
              - child_group1
    ```
    """

    def __init__(
        self,
        config: Config,
        parallelize_platforms: bool = False,  # noqa: FBT001, FBT002
        platform_name: str | None = None,
    ) -> None:
        """Initialize a new platform class and returns None.

        Args:
            config: An instance of a Molecule config.
            parallelize_platforms: Parallel mode. Default is False.
            platform_name: One platform to target only, defaults to None.
        """
        if platform_name:
            config.config["platforms"] = util._filter_platforms(  # noqa: SLF001
                config.config,
                platform_name,
            )
        if parallelize_platforms:
            config.config["platforms"] = util._parallelize_platforms(  # noqa: SLF001
                config.config,
                config._run_uuid,  # noqa: SLF001
            )
        self._config = config

    @property
    def instances(self) -> list[PlatformData]:
        """The platforms defined in the config.

        Returns:
            The list of platforms in the config.
        """
        return self._config.config["platforms"]



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Ansible-Playbook Provisioner Module."""

from __future__ import annotations

import shlex
import subprocess
import warnings

from typing import TYPE_CHECKING

from rich.markup import escape

from molecule import logger, util
from molecule.api import MoleculeRuntimeWarning
from molecule.exceptions import ScenarioFailureError
from molecule.types import ScenarioResult


if TYPE_CHECKING:
    from molecule.config import Config


class AnsiblePlaybook:
    """Provisioner Playbook."""

    def __init__(
        self,
        playbook: str | None,
        config: Config,
        *,
        verify: bool = False,
    ) -> None:
        """Set up the requirements to execute ``ansible-playbook``.

        Args:
            playbook: A string containing the path to the playbook.
            config: An instance of a Molecule config.
            verify: An optional bool to toggle the Playbook mode between provision and verify.
                False provision; True: verify. Default is False.
        """
        self._ansible_command: list[str] = []
        self._playbook = playbook
        self._config = config
        self._cli: dict[str, str | bool] = {}
        self._env: dict[str, str] = {}
        if verify:
            self._env = util.merge_dicts(
                self._config.verifier.env,
                self._config.config["verifier"]["env"],
            )
        elif self._config.provisioner:
            self._env = self._config.provisioner.env

    @property
    def _log(self) -> logger.ScenarioLoggerAdapter:
        """Get a fresh scenario logger with current context.

        Returns:
            A scenario logger adapter with current scenario and step context.
        """
        # Get step context from the current action being executed
        step_name = getattr(self._config, "action", "provisioner")
        return logger.get_scenario_logger(__name__, self._config.scenario.name, step_name)

    def bake(self) -> None:
        """Bake ``ansible-playbook`` or ``navigator run`` command so it's ready to execute.

        Raises:
            ValueError: when backend is incorrect.
            RuntimeError: when ansible-playbook or ansible-navigator is not available.
        """
        if not self._playbook:
            return

        if self._config.provisioner:
            # Pass a directory as inventory to let Ansible merge the multiple
            # inventory sources located under
            self.add_cli_arg("inventory", self._config.provisioner.inventory_directory)
            options = util.merge_dicts(self._config.provisioner.options, self._cli)
            verbose_flag = util.verbose_flag(options)
            if self._playbook != self._config.provisioner.playbooks.converge:  # noqa: SIM102
                if options.get("become"):
                    del options["become"]

            # We do not pass user-specified Ansible arguments to the create and
            # destroy invocations because playbooks involved in those two
            # operations are not always provided by end users. And in those cases,
            # custom Ansible arguments can break the creation and destruction
            # processes.
            #
            # If users need to modify the creation of deletion, they can supply
            # custom playbooks and specify them in the scenario configuration.
            if self._config.action not in ["create", "destroy"]:
                ansible_args = list(self._config.provisioner.ansible_args) + list(
                    self._config.ansible_args,
                )
            else:
                ansible_args = []

            backend = self._config.executor

            if backend:
                try:
                    result = subprocess.run(
                        [backend, "--version"],
                        capture_output=True,
                        text=True,
                        check=True,
                    )
                    self._log.info("%s version: %s", backend, result.stdout.strip())
                except subprocess.CalledProcessError as exc:
                    msg = f"{backend} is not available. Please ensure that it is installed."
                    raise RuntimeError(msg) from exc

            if backend == "ansible-playbook":
                self._ansible_command = [
                    "ansible-playbook",
                    *util.dict2args(options),
                    *util.bool2args(verbose_flag),
                    *ansible_args,
                    self._playbook,  # must always go last
                ]

            elif backend == "ansible-navigator":
                self._ansible_command = [
                    "ansible-navigator",
                    "run",
                    self._playbook,
                    "--mode",
                    "stdout",
                    *util.dict2args(options),
                    *util.bool2args(verbose_flag),
                    *ansible_args,
                ]
            else:
                msg = f"Unsupported backend: {backend}"
                raise ValueError(msg)

    def execute(self, action_args: list[str] | None = None) -> str:  # noqa: ARG002
        """Execute ``ansible-playbook`` or ``ansible-navigator run``.

        Args:
            action_args: Arguments to forward to the action. Unused.

        Returns:
            Output from ansible-playbook or ansible-navigator.

        Raises:
            ScenarioFailureError: when Ansible returns nonzero code.
        """
        if not self._ansible_command:
            self.bake()

        if not self._playbook:
            self._log.warning("Skipping, %s action has no playbook.", self._config.action)
            self._config.scenario.results.append(
                ScenarioResult(subcommand=self._config.action, state="SKIPPED"),
            )
            return ""

        with warnings.catch_warnings(record=True) as warns:
            warnings.filterwarnings("default", category=MoleculeRuntimeWarning)
            self._config.driver.sanity_checks()
            cwd = self._config.scenario_path
            result = self._config.app.run_command(
                cmd=self._ansible_command,
                env=self._env,
                debug=self._config.debug,
                cwd=cwd,
            )

        if result.returncode != 0:
            self._config.scenario.results.append(
                ScenarioResult(subcommand=self._config.action, state="FAILED"),
            )

            msg = f"Ansible return code was {result.returncode}, command was: [dim]{escape(shlex.join(result.args))}[/dim]"
            raise ScenarioFailureError(
                msg,
                code=result.returncode,
                warns=warns,
            )

        self._config.scenario.results.append(
            ScenarioResult(subcommand=self._config.action, state="PASSED"),
        )
        return result.stdout

    def add_cli_arg(self, name: str, value: str | bool) -> None:  # noqa: FBT001
        """Add argument to CLI passed to ansible-playbook.

        Args:
            name: A string containing the name of argument to be added.
            value: The value of argument to be added.
        """
        if value:
            self._cli[name] = value

    def add_env_arg(self, name: str, value: str) -> None:
        """Add argument to environment passed to ansible-playbook.

        Args:
            name: A string containing the name of argument to be added.
            value: The value of argument to be added.
        """
        self._env[name] = value



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Base Dependency Module."""

from __future__ import annotations

import abc
import os
import time

from subprocess import CalledProcessError
from typing import TYPE_CHECKING

from molecule import logger, util


if TYPE_CHECKING:
    from collections.abc import MutableMapping

    from molecule.config import Config


class Base(abc.ABC):
    """Dependency Base Class.

    Attributes:
        RETRY: Number of times to retry the dependency.
        SLEEP: Number of seconds to sleep between retries.
        BACKOFF: Additional number of seconds to sleep for each successive attempt.
    """

    RETRY = 3
    SLEEP = 3
    BACKOFF = 3

    def __init__(self, config: Config) -> None:
        """Initialize code for all :ref:`Dependency` classes.

        Args:
            config: An instance of a Molecule config.
        """
        self._config = config
        self._sh_command: str | list[str] = []

    @property
    def _log(self) -> logger.ScenarioLoggerAdapter:
        """Get a fresh scenario logger with current context.

        Returns:
            A scenario logger adapter with current scenario and step context.
        """
        # Get step context from the current action being executed
        step_name = getattr(self._config, "action", "dependency")
        return logger.get_scenario_logger(__name__, self._config.scenario.name, step_name)

    def execute_with_retries(self) -> None:
        """Run dependency downloads with retry and timed back-off."""
        try:
            self._config.app.run_command(self._sh_command, debug=self._config.debug, check=True)
            msg = "Dependency completed successfully."
            self._log.info(msg)
            return  # noqa: TRY300
        except CalledProcessError:
            pass

        for counter in range(1, (self.RETRY + 1)):
            msg = f"Retrying dependency ... {counter:d}/{self.RETRY:d} time(s)"
            self._log.warning(msg)

            msg = f"Sleeping for {self.SLEEP:d} seconds before retrying ..."
            self._log.warning(msg)
            time.sleep(self.SLEEP)
            self.SLEEP += self.BACKOFF

            try:
                self._config.app.run_command(self._sh_command, debug=self._config.debug, check=True)
                msg = "Dependency completed successfully."
                self._log.info(msg)
                return  # noqa: TRY300
            except CalledProcessError as _exception:
                exception = _exception

        self._log.error(str(exception))
        util.sysexit(exception.returncode)

    @abc.abstractmethod
    def execute(
        self,
        action_args: list[str] | None = None,
    ) -> None:  # pragma: no cover
        """Execute ``cmd``.

        Args:
            action_args: Arguments for dependency resolvers. Unused.
        """
        for name, version in self._config.driver.required_collections.items():
            self._config.runtime.require_collection(name, version)

    @property
    @abc.abstractmethod
    def default_options(self) -> MutableMapping[str, str | bool]:  # pragma: no cover
        """Get default CLI arguments provided to ``cmd``.

        Returns:
            dict
        """

    @property
    def default_env(self) -> dict[str, str]:  # pragma: no cover
        """Get default env variables provided to ``cmd``.

        Returns:
            dict
        """
        # I do not know why mypy has a problem here. We are merging two dict[str, str]s into one.
        # dict[str, str] should fit the typevar of merge_dicts, and all types are the same, yet
        # it still complains.
        env = dict(os.environ)
        return util.merge_dicts(env, self._config.env)

    @property
    def name(self) -> str:
        """Name of the dependency and returns a string.

        :returns: str
        """
        return self._config.config["dependency"]["name"]

    @property
    def enabled(self) -> bool:
        """Is the dependency enabled.

        Returns:
            Whether the dependency is enabled.
        """
        return self._config.config["dependency"]["enabled"]

    @property
    def options(self) -> MutableMapping[str, str | bool]:
        """Computed dependency options.

        Returns:
            Merged dictionary of default options with config options.
        """
        return util.merge_dicts(
            self.default_options,
            self._config.config["dependency"]["options"],
        )

    @property
    def env(self) -> dict[str, str]:
        """Computed environment variables.

        Returns:
            Merged dictionary of default env vars with config env vars.
        """
        return util.merge_dicts(
            self.default_env,
            self._config.config["dependency"]["env"],
        )



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Converge Command Module."""

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from molecule.click_cfg import (
    click_command_ex,
    common_options,
)
from molecule.command import base


if TYPE_CHECKING:
    import click

    from molecule.types import CommandArgs, MoleculeArgs


class Converge(base.Base):
    """Converge Command Class."""

    def execute(self, action_args: list[str] | None = None) -> None:  # noqa: ARG002
        """Execute the actions necessary to perform a `molecule converge`.

        Args:
            action_args: Arguments for this command. Unused.
        """
        if self._config.provisioner:
            self._config.provisioner.converge()
        self._config.state.change_state("converged", value=True)


@click_command_ex()
@common_options("ansible_args")
def converge(ctx: click.Context) -> None:  # pragma: no cover
    """Use the provisioner to configure instances (dependency, create, prepare converge).

    Args:
        ctx: Click context object holding commandline arguments.
    """
    args: MoleculeArgs = ctx.obj.get("args")
    subcommand = base._get_subcommand(__name__)  # noqa: SLF001
    command_args: CommandArgs = {
        "report": ctx.params["report"],
        "shared_inventory": ctx.params["shared_inventory"],
        "shared_state": ctx.params["shared_state"],
        "subcommand": subcommand,
    }

    __all = ctx.params["all"]
    ansible_args = ctx.params["ansible_args"]
    exclude = ctx.params["exclude"]
    scenario_name = ctx.params["scenario_name"]

    if __all:
        scenario_name = None

    base.execute_cmdline_scenarios(scenario_name, args, command_args, ansible_args, exclude)



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import pytest

from molecule.model import schema_v3


@pytest.fixture
def _model_dependency_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {
        "dependency": {
            "name": "galaxy",
            "enabled": True,
            "options": {"foo": "bar"},
            "env": {"FOO": "foo", "FOO_BAR": "foo_bar"},
        },
    }


@pytest.mark.parametrize(
    "config",
    ["_model_dependency_section_data"],  # noqa: PT007
    indirect=True,
)
def test_dependency(config):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    assert not schema_v3.validate(config)


@pytest.fixture
def _model_dependency_errors_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"dependency": {"name": 0}}


@pytest.mark.parametrize(
    "config",
    ["_model_dependency_errors_section_data"],  # noqa: PT007
    indirect=True,
)
def test_dependency_has_errors(config):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = ["0 is not one of ['galaxy', 'shell']"]

    assert x == schema_v3.validate(config)


@pytest.fixture
def _model_dependency_allows_galaxy_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"dependency": {"name": "galaxy"}}


@pytest.fixture
def _model_dependency_allows_shell_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"dependency": {"name": "shell"}}


@pytest.mark.parametrize(
    "config",
    [  # noqa: PT007
        ("_model_dependency_allows_galaxy_section_data"),
        ("_model_dependency_allows_shell_section_data"),
    ],
    indirect=True,
)
def test_dependency_allows_shell_name(config):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    assert not schema_v3.validate(config)


@pytest.fixture
def _model_dependency_shell_errors_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"dependency": {"name": "shell", "command": None}}


@pytest.mark.parametrize(
    "config",
    ["_model_dependency_shell_errors_section_data"],  # noqa: PT007
    indirect=True,
)
def test_dependency_shell_has_errors(config):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = ["None is not of type 'string'"]

    assert x == schema_v3.validate(config)



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import typing

from molecule.model import schema_v3
from tests.conftest import run


if typing.TYPE_CHECKING:
    from pathlib import Path


def test_base_config(config):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    assert not schema_v3.validate(config)


def test_molecule_schema(resources_folder_path: Path) -> None:
    """Test the molecule schema.

    Args:
        resources_folder_path: Path to the resources folder.
    """
    cmd = [
        "uv",
        "tool",
        "run",
        "check-jsonschema",
        "-v",
        "--schemafile",
        "src/molecule/data/molecule.json",
        f"{resources_folder_path}/schema_instance_files/valid/molecule.yml",
    ]
    assert run(cmd).returncode == 0

    cmd = [
        "uv",
        "tool",
        "run",
        "check-jsonschema",
        "-v",
        "--schemafile",
        "src/molecule/data/driver.json",
        f"{resources_folder_path}/schema_instance_files/invalid/molecule_delegated.yml",
    ]
    assert run(cmd).returncode != 0



"""Tests for the new Click configuration system with CliOption architecture."""

from __future__ import annotations

from dataclasses import replace

import click
import pytest

from click.testing import CliRunner

from molecule.click_cfg import CliOption, CliOptions, _sort_options, common_options


def test_basic_option_creation() -> None:
    """Test creating a basic CliOption."""
    option = CliOption(
        name="test-option",
        help="Test help text",
        short="-t",
        default="default_value",
    )

    assert option.name == "test-option"
    assert option.help == "Test help text"
    assert option.short == "-t"
    assert option.default == "default_value"
    assert option.multiple is False
    assert option.is_flag is False


def test_flag_option_creation() -> None:
    """Test creating a flag option."""
    option = CliOption(
        name="verbose",
        help="Enable verbose output",
        is_flag=True,
        default=False,
    )

    assert option.is_flag is True
    assert option.default is False


def test_multiple_option_creation() -> None:
    """Test creating a multiple option."""
    option = CliOption(
        name="tags",
        help="Specify tags",
        multiple=True,
        short="-t",
    )

    assert option.multiple is True


def test_argument_option_creation() -> None:
    """Test creating an argument option."""
    option = CliOption(
        name="ansible_args",
        help="Ansible arguments",
        is_argument=True,
        nargs=-1,
    )

    assert option.is_argument is True
    assert option.nargs == -1


def test_as_click_option_basic() -> None:
    """Test converting basic option to Click option."""
    option = CliOption(
        name="test-option",
        help="Test help",
        short="-t",
        default="value",
    )

    click_option = option.as_click_option()

    # Apply to a dummy function to test
    @click_option
    def dummy() -> None:
        pass

    assert hasattr(dummy, "__click_params__")
    param = dummy.__click_params__[0]
    assert param.opts == ["--test-option", "-t"]
    assert param.help == "Test help (default: value)"  # Now expect generated help text


def test_as_click_option_flag() -> None:
    """Test converting flag option to Click option."""
    option = CliOption(
        name="verbose",
        help="Enable verbose output",
        is_flag=True,
        default=False,
    )

    click_option = option.as_click_option()

    # Apply to a dummy function to test
    @click_option
    def dummy() -> None:
        pass

    assert hasattr(dummy, "__click_params__")
    param = dummy.__click_params__[0]
    assert param.opts == ["--verbose"]
    assert param.secondary_opts == ["--no-verbose"]
    assert param.help == "Enable verbose output (default: disabled)"


def test_as_click_option_with_choices() -> None:
    """Test converting option with choices to Click option."""
    option = CliOption(
        name="format",
        help="Output format",
        choices=["json", "yaml"],
        default="json",
    )

    click_option = option.as_click_option()

    # Apply to a dummy function to test
    @click_option
    def dummy() -> None:
        pass

    assert hasattr(dummy, "__click_params__")
    param = dummy.__click_params__[0]
    assert param.help == "Output format (default: json)"  # Expect generated help text


def test_as_click_option_multiple() -> None:
    """Test converting multiple option to Click option."""
    option = CliOption(
        name="tags",
        help="Specify tags",
        multiple=True,
        short="-t",
    )

    click_option = option.as_click_option()

    @click_option
    def dummy() -> None:
        pass

    param = dummy.__click_params__[0]
    assert param.multiple is True


def test_as_click_option_argument() -> None:
    """Test converting argument to Click argument."""
    option = CliOption(
        name="ansible_args",
        help="Ansible arguments",
        is_argument=True,
        nargs=-1,
    )

    click_option = option.as_click_option()

    @click_option
    def dummy() -> None:
        pass

    param = dummy.__click_params__[0]
    assert param.name == "ansible_args"
    assert param.nargs == -1
    assert param.type == click.UNPROCESSED


def test_replace_functionality() -> None:
    """Test that dataclass replace works correctly."""
    base_option = CliOption(
        name="scenario-name",
        help="Base scenario option",
        short="-s",
        multiple=True,
    )

    modified_option = replace(base_option, default=["default"], multiple=False)

    assert modified_option.name == "scenario-name"
    assert modified_option.help == "Base scenario option"
    assert modified_option.short == "-s"
    assert modified_option.default == ["default"]
    assert modified_option.multiple is False


def test_help_text_generation_basic() -> None:
    """Test basic help text generation."""
    option = CliOption(
        name="test-option",
        help="Test option",
        default="test_value",
    )

    help_text = option._generate_help_text()
    assert help_text == "Test option (default: test_value)"


def test_help_text_generation_experimental() -> None:
    """Test experimental help text generation."""
    option = CliOption(
        name="test-option",
        help="Test option",
        experimental=True,
    )

    help_text = option._generate_help_text()
    assert help_text == "EXPERIMENTAL: Test option"


def test_help_text_generation_experimental_with_default() -> None:
    """Test experimental help text with default value."""
    option = CliOption(
        name="test-option",
        help="Test option",
        default="test_value",
        experimental=True,
    )

    help_text = option._generate_help_text()
    assert help_text == "EXPERIMENTAL: Test option (default: test_value)"


def test_help_text_generation_flag_enabled() -> None:
    """Test that flag options show enabled status correctly."""
    option = CliOption(
        name="test-flag",
        help="Test flag",
        is_flag=True,
        default=True,
    )

    help_text = option._generate_help_text()
    assert help_text == "Test flag (default: enabled)"


def test_help_text_generation_flag_disabled() -> None:
    """Test that flag options show disabled status correctly."""
    option = CliOption(
        name="test-flag",
        help="Test flag",
        is_flag=True,
        default=False,
    )

    help_text = option._generate_help_text()
    assert help_text == "Test flag (default: disabled)"


def test_help_text_generation_custom_default() -> None:
    """Test custom help_default overrides automatic generation."""
    option = CliOption(
        name="test-option",
        help="Test option",
        default="actual_value",
        help_default="custom_value",
    )

    help_text = option._generate_help_text()
    assert help_text == "Test option (default: custom_value)"


def test_help_text_generation_custom_default_flag() -> None:
    """Test custom help_default overrides flag automatic generation."""
    option = CliOption(
        name="test-flag",
        help="Test flag",
        is_flag=True,
        default=False,
        help_default="custom_status",
    )

    help_text = option._generate_help_text()
    assert help_text == "Test flag (default: custom_status)"


def test_help_text_generation_experimental_custom_default() -> None:
    """Test experimental with custom help_default."""
    option = CliOption(
        name="test-option",
        help="Test option",
        experimental=True,
        help_default="custom_value",
    )

    help_text = option._generate_help_text()
    assert help_text == "EXPERIMENTAL: Test option (default: custom_value)"


def test_help_text_generation_multiple_list_default() -> None:
    """Test multiple option with list default."""
    option = CliOption(
        name="test-multiple",
        help="Test multiple",
        multiple=True,
        default=["value1", "value2"],
    )

    help_text = option._generate_help_text()
    assert help_text == "Test multiple (default: value1, value2)"


def test_help_text_generation_multiple_single_default() -> None:
    """Test multiple option with single item list default."""
    option = CliOption(
        name="test-multiple",
        help="Test multiple",
        multiple=True,
        default=["value1"],
    )

    help_text = option._generate_help_text()
    assert help_text == "Test multiple (default: value1)"


def test_help_text_generation_argument_no_default() -> None:
    """Test that arguments don't show default info."""
    option = CliOption(
        name="test-arg",
        help="Test argument",
        is_argument=True,
        default="value",
    )

    help_text = option._generate_help_text()
    assert help_text == "Test argument"
    assert "default" not in help_text


def test_help_text_generation_choice_with_default() -> None:
    """Test choice option with default."""
    option = CliOption(
        name="test-choice",
        help="Test choice",
        choices=["option1", "option2"],
        default="option1",
    )

    help_text = option._generate_help_text()
    assert help_text == "Test choice (default: option1)"


def test_cli_options_instantiation() -> None:
    """Test that CliOptions can be instantiated."""
    options = CliOptions()
    assert isinstance(options, CliOptions)


def test_scenario_name_base() -> None:
    """Test the base scenario_name option."""
    options = CliOptions()
    scenario = options.scenario_name

    assert scenario.name == "scenario-name"
    assert scenario.short == "-s"
    assert scenario.multiple is True
    assert scenario.default is None


def test_scenario_name_with_default() -> None:
    """Test scenario_name with default."""
    options = CliOptions()
    scenario = options.scenario_name_with_default

    assert scenario.name == "scenario-name"
    assert scenario.default == ["default"]  # MOLECULE_DEFAULT_SCENARIO_NAME as list
    assert "default" in scenario._generate_help_text()  # Check generated help text


def test_scenario_name_single() -> None:
    """Test single scenario_name without default."""
    options = CliOptions()
    scenario = options.scenario_name_single

    assert scenario.name == "scenario-name"
    assert scenario.multiple is False
    assert scenario.default is None


def test_scenario_name_single_with_default() -> None:
    """Test single scenario_name with default."""
    options = CliOptions()
    scenario = options.scenario_name_single_with_default

    assert scenario.name == "scenario-name"
    assert scenario.multiple is False
    assert scenario.default == "default"
    assert "default" in scenario._generate_help_text()  # Check generated help text


def test_exclude_option() -> None:
    """Test the exclude option."""
    options = CliOptions()
    exclude = options.exclude

    assert exclude.name == "exclude"
    assert exclude.short == "-e"
    assert exclude.multiple is True


def test_all_scenarios_option() -> None:
    """Test the all_scenarios flag option."""
    options = CliOptions()
    all_opt = options.all_scenarios

    assert all_opt.name == "all"
    assert all_opt.is_flag is True
    assert all_opt.default is False


def test_driver_options() -> None:
    """Test driver-related options."""
    options = CliOptions()

    driver = options.driver_name
    assert driver.name == "driver-name"
    assert driver.short == "-d"

    driver_with_choices = options.driver_name_with_choices
    assert driver_with_choices.name == "driver-name"
    assert driver_with_choices.choices is not None
    assert len(driver_with_choices.choices) > 0  # Should have available drivers


def test_platform_options() -> None:
    """Test platform-related options."""
    options = CliOptions()

    platform = options.platform_name
    assert platform.name == "platform-name"
    assert platform.short == "-p"

    platform_with_default = options.platform_name_with_default
    assert platform_with_default.default is None  # MOLECULE_PLATFORM_NAME defaults to None
    assert platform_with_default.help_default == "None"
    assert "(default: None)" in platform_with_default._generate_help_text()


def test_execution_options() -> None:
    """Test execution-related options."""
    options = CliOptions()

    parallel = options.parallel
    assert parallel.name == "parallel"
    assert parallel.is_flag is True
    assert "(default: disabled)" in parallel._generate_help_text()

    force = options.force
    assert force.name == "force"
    assert force.short == "-f"
    assert force.is_flag is True
    assert "(default: disabled)" in force._generate_help_text()

    destroy = options.destroy
    assert destroy.name == "destroy"
    assert destroy.choices == ["always", "never"]
    assert destroy.default == "always"
    assert "(default: always)" in destroy._generate_help_text()


def test_output_options() -> None:
    """Test output and reporting options."""
    options = CliOptions()

    report = options.report
    assert report.name == "report"
    assert report.is_flag is True
    assert report.experimental is True
    help_text = report._generate_help_text()
    assert "EXPERIMENTAL:" in help_text
    assert "(default: disabled)" in help_text

    shared_inventory = options.shared_inventory
    assert shared_inventory.name == "shared-inventory"
    assert shared_inventory.is_flag is True
    assert shared_inventory.experimental is True
    help_text = shared_inventory._generate_help_text()
    assert "EXPERIMENTAL:" in help_text
    assert "(default: disabled)" in help_text

    shared_state = options.shared_state
    assert shared_state.name == "shared-state"
    assert shared_state.is_flag is True
    assert shared_state.experimental is True
    help_text = shared_state._generate_help_text()
    assert "EXPERIMENTAL:" in help_text
    assert "(default: disabled)" in help_text


def test_format_options() -> None:
    """Test format options."""
    options = CliOptions()

    format_simple = options.format_simple
    assert format_simple.name == "format"
    assert format_simple.short == "-f"
    assert format_simple.choices == ["simple", "plain"]
    assert "(default: simple)" in format_simple._generate_help_text()

    format_full = options.format_full
    assert format_full.choices == ["simple", "plain", "yaml"]
    assert "(default: simple)" in format_full._generate_help_text()


def test_connection_options() -> None:
    """Test connection options."""
    options = CliOptions()

    host = options.host
    assert host.name == "host"
    assert host.short == "-h"


def test_dependency_provisioner_options() -> None:
    """Test dependency and provisioner options."""
    options = CliOptions()

    dependency = options.dependency_name
    assert dependency.name == "dependency-name"
    assert dependency.choices == ["galaxy"]
    assert dependency.default == "galaxy"
    assert "(default: galaxy)" in dependency._generate_help_text()

    provisioner = options.provisioner_name
    assert provisioner.name == "provisioner-name"
    assert provisioner.choices == ["ansible"]
    assert provisioner.default == "ansible"
    assert "(default: ansible)" in provisioner._generate_help_text()


def test_argument_options() -> None:
    """Test argument options."""
    options = CliOptions()

    ansible_args = options.ansible_args
    assert ansible_args.name == "ansible_args"
    assert ansible_args.is_argument is True
    assert ansible_args.nargs == -1

    subcommand = options.subcommand
    assert subcommand.name == "subcommand"
    assert subcommand.is_argument is True
    assert subcommand.nargs == 1


def test_experimental_flag_functionality() -> None:
    """Test that experimental flag works correctly."""
    options = CliOptions()

    # Test experimental options
    experimental_options = [
        options.report,
        options.shared_inventory,
        options.shared_state,
    ]

    for option in experimental_options:
        assert option.experimental is True
        help_text = option._generate_help_text()
        assert help_text.startswith("EXPERIMENTAL:")

    # Test non-experimental options
    non_experimental_options = [
        options.parallel,
        options.force,
        options.destroy,
    ]

    for option in non_experimental_options:
        assert option.experimental is False
        help_text = option._generate_help_text()
        assert not help_text.startswith("EXPERIMENTAL:")


def test_help_text_generation() -> None:
    """Test automatic help text generation with defaults."""
    options = CliOptions()

    # Test flag option (should show disabled/enabled)
    force = options.force
    force_help = force._generate_help_text()
    assert "(default: disabled)" in force_help

    # Test choice option with default
    destroy = options.destroy
    destroy_help = destroy._generate_help_text()
    assert "(default: always)" in destroy_help

    # Test multiple option with list default
    scenario = options.scenario_name_with_default
    scenario_help = scenario._generate_help_text()
    assert "(default: default)" in scenario_help

    # Test option without default
    host = options.host
    host_help = host._generate_help_text()
    assert "default" not in host_help

    # Test argument (should not show default info)
    ansible_args = options.ansible_args
    args_help = ansible_args._generate_help_text()
    assert "default" not in args_help

    # Test custom help_default
    platform_with_default = options.platform_name_with_default
    platform_help = platform_with_default._generate_help_text()
    assert "(default: None)" in platform_help


def test_decorator_with_no_additional_options() -> None:
    """Test decorator with no additional options (just COMMON_OPTIONS)."""

    @click.command()
    @common_options()
    def test_command(ctx: click.Context) -> None:
        # Should have all COMMON_OPTIONS available
        scenario = ctx.params["scenario_name"]
        exclude = ctx.params["exclude"]
        report = ctx.params["report"]
        click.echo(f"scenario={scenario}, exclude={exclude}, report={report}")

    runner = CliRunner()
    result = runner.invoke(
        test_command,
        ["--scenario-name", "test", "--exclude", "skip", "--report"],
    )

    assert result.exit_code == 0
    assert "scenario=('test',)" in result.output
    assert "exclude=('skip',)" in result.output
    assert "report=True" in result.output


def test_decorator_with_additional_string_options() -> None:
    """Test decorator with additional string options."""

    @click.command()
    @common_options("parallel", "force")
    def test_command(ctx: click.Context) -> None:
        # Should have COMMON_OPTIONS plus additional ones
        scenario = ctx.params["scenario_name"]
        parallel = ctx.params["parallel"]
        force = ctx.params["force"]
        click.echo(f"scenario={scenario}, parallel={parallel}, force={force}")

    runner = CliRunner()
    result = runner.invoke(test_command, ["--scenario-name", "test", "--parallel", "--force"])

    assert result.exit_code == 0
    assert "scenario=('test',)" in result.output
    assert "parallel=True" in result.output
    assert "force=True" in result.output


def test_decorator_deduplication() -> None:
    """Test that decorator deduplicates options when same option is in COMMON_OPTIONS and additional."""

    @click.command()
    @common_options(
        "scenario_name_with_default",
        "report",
    )  # report is already in COMMON_OPTIONS
    def test_command(ctx: click.Context) -> None:
        scenario = ctx.params["scenario_name"]
        report = ctx.params["report"]
        click.echo(f"scenario={scenario}, report={report}")

    runner = CliRunner()
    result = runner.invoke(test_command, ["--scenario-name", "test", "--report"])

    assert result.exit_code == 0
    assert "scenario=('test',)" in result.output
    assert "report=True" in result.output


def test_decorator_with_single_option() -> None:
    """Test decorator with a single option."""

    @click.command()  # Applied last - converts to Click command
    @common_options("parallel")  # Applied first - adds options
    def test_command(ctx: click.Context) -> None:
        click.echo(f"parallel={ctx.params['parallel']}")

    runner = CliRunner()
    result = runner.invoke(test_command, ["--parallel"])

    assert result.exit_code == 0
    assert "parallel=True" in result.output


def test_decorator_with_multiple_options() -> None:
    """Test decorator with multiple options."""

    @click.command()
    @common_options("scenario_name_with_default", "parallel", "force")
    def test_command(ctx: click.Context) -> None:
        scenario = ctx.params["scenario_name"]
        parallel = ctx.params["parallel"]
        force = ctx.params["force"]
        click.echo(f"scenario={scenario}, parallel={parallel}, force={force}")

    runner = CliRunner()
    result = runner.invoke(test_command, ["--scenario-name", "test", "--parallel", "--force"])

    assert result.exit_code == 0
    assert "scenario=('test',)" in result.output  # Fixed: tuple format
    assert "parallel=True" in result.output
    assert "force=True" in result.output


def test_decorator_with_argument() -> None:
    """Test decorator with argument option."""

    @click.command()
    @common_options("ansible_args")
    def test_command(ctx: click.Context) -> None:
        args = ctx.params["ansible_args"]
        click.echo(f"args={args}")

    runner = CliRunner()
    result = runner.invoke(test_command, ["--", "--extra-vars", "test=1"])

    assert result.exit_code == 0
    assert "args=('--extra-vars', 'test=1')" in result.output


def test_decorator_with_choices() -> None:
    """Test decorator with choice option."""

    @click.command()
    @common_options("format_simple")
    def test_command(ctx: click.Context) -> None:
        output_format = ctx.params["format"]
        click.echo(f"format={output_format}")

    runner = CliRunner()
    result = runner.invoke(test_command, ["--format", "plain"])

    assert result.exit_code == 0
    assert "format=plain" in result.output


def test_decorator_with_multiple_scenarios() -> None:
    """Test decorator with multiple scenario names."""

    @click.command()
    @common_options("scenario_name")
    def test_command(ctx: click.Context) -> None:
        scenarios = ctx.params["scenario_name"]
        click.echo(f"scenarios={scenarios}")

    runner = CliRunner()
    result = runner.invoke(
        test_command,
        ["--scenario-name", "test1", "--scenario-name", "test2"],
    )

    assert result.exit_code == 0
    assert "scenarios=('test1', 'test2')" in result.output  # Fixed: tuple format


def test_decorator_help_output() -> None:
    """Test that help output includes all options."""

    @click.command()
    @common_options("scenario_name_with_default", "parallel")
    def test_command(ctx: click.Context) -> None:
        pass

    runner = CliRunner()
    result = runner.invoke(test_command, ["--help"])

    assert result.exit_code == 0
    assert "--scenario-name" in result.output
    assert "--parallel" in result.output
    assert "-s" in result.output  # short form for scenario


def test_decorator_invalid_option_name() -> None:
    """Test decorator with invalid option name."""
    with pytest.raises(AttributeError):

        @click.command()
        @common_options("nonexistent_option")
        def test_command(ctx: click.Context) -> None:
            pass


def test_ctx_only_function_signature() -> None:
    """Test that functions only need ctx parameter."""

    @click.command()
    @common_options("parallel", "force")
    def test_command(ctx: click.Context) -> None:
        # Function signature only has ctx, but can access all options
        parallel = ctx.params["parallel"]
        force = ctx.params["force"]
        click.echo(f"parallel={parallel}, force={force}")

    runner = CliRunner()
    result = runner.invoke(test_command, ["--parallel", "--force"])

    assert result.exit_code == 0
    assert "parallel=True" in result.output
    assert "force=True" in result.output


def test_complex_scenario_composition() -> None:
    """Test complex scenario with multiple option types."""

    @click.command()
    @common_options(
        "scenario_name_with_default",
        "exclude",
        "all_scenarios",
        "driver_name_with_choices",
        "parallel",
        "format_simple",
    )
    def test_command(ctx: click.Context) -> None:
        params = ctx.params
        click.echo(f"Complex command: {params}")

    runner = CliRunner()
    result = runner.invoke(
        test_command,
        [
            "--scenario-name",
            "test1",
            "--exclude",
            "excluded",
            "--all",
            "--driver-name",
            "default",  # Fixed: use 'default' instead of 'delegated'
            "--parallel",
            "--format",
            "plain",
        ],
    )

    assert result.exit_code == 0
    # Verify all parameters are accessible
    assert "scenario_name" in result.output
    assert "exclude" in result.output
    assert "all" in result.output
    assert "driver_name" in result.output
    assert "parallel" in result.output
    assert "format" in result.output


def test_option_sort_order() -> None:
    """Test that options are sorted in the correct order."""
    cli_options = CliOptions()

    # Create a test set with options from each category
    test_options = [
        cli_options.force,  # Short form, non-experimental
        cli_options.scenario_name_with_default,  # Should be first (scenario-name)
        cli_options.report,  # Experimental
        cli_options.exclude,  # Should be second
        cli_options.all_scenarios,  # Should be third (all)
        cli_options.parallel,  # Long form, non-experimental
        cli_options.shared_state,  # Experimental
        cli_options.driver_name,  # Short form, non-experimental
    ]

    sorted_options = _sort_options(test_options)
    sorted_names = [opt.name for opt in sorted_options]

    # Verify the expected order
    expected_order = [
        "scenario-name",  # Section 1: Core workflow (scenario-name, exclude, all)
        "exclude",  # Section 1: Core workflow (scenario-name, exclude, all)
        "all",  # Section 1: Core workflow (scenario-name, exclude, all)
        "driver-name",  # Section 2: short forms (alphabetical): driver-name
        "force",  # Section 2: short forms (alphabetical): force
        "parallel",  # Section 3: long forms (alphabetical)
        "report",  # Section 4: experimental (alphabetical): report
        "shared-state",  # Section 4: experimental (alphabetical): shared-state
    ]

    assert sorted_names == expected_order, f"Expected {expected_order}, got {sorted_names}"



"""Unit tests for the ansible verifier module."""

from __future__ import annotations

import os

import pytest

from molecule import config
from molecule.verifier import ansible


@pytest.fixture
def _patched_ansible_verify(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN202
    m = mocker.patch("molecule.provisioner.ansible.Ansible.verify")
    m.return_value = "patched-ansible-verify-stdout"

    return m


@pytest.fixture
def _verifier_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"verifier": {"name": "ansible", "env": {"FOO": "bar"}}}


# NOTE(retr0h): The use of the `patched_config_validate` fixture, disables
# config.Config._validate from executing.  Thus preventing odd side-effects
# throughout patched.assert_called unit tests.
@pytest.fixture
def _instance(  # type: ignore[no-untyped-def]  # noqa: ANN202
    _verifier_section_data,
    patched_config_validate,
    config_instance: config.Config,
):
    return ansible.Ansible(config_instance)


def test_verifier_config_private_member(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert isinstance(_instance._config, config.Config)


def test_verifier_default_options_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.default_options == {}


def test_verifier_ansible_default_env_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert "MOLECULE_FILE" in _instance.default_env
    assert "MOLECULE_INVENTORY_FILE" in _instance.default_env
    assert "MOLECULE_SCENARIO_DIRECTORY" in _instance.default_env
    assert "MOLECULE_INSTANCE_CONFIG" in _instance.default_env


@pytest.mark.parametrize(
    "config_instance",
    ["_verifier_section_data"],  # noqa: PT007
    indirect=True,
)
def test_verifier_env_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.env["FOO"] == "bar"


def test_verifier_name_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.name == "ansible"


def test_ansible_enabled_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.enabled


def test_verifier_directory_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    parts = _instance.directory.split(os.path.sep)
    # Unused by Ansible verifier
    assert parts[-3:] == ["molecule", "default", "tests"]


@pytest.mark.parametrize(
    "config_instance",
    ["_verifier_section_data"],  # noqa: PT007
    indirect=True,
)
def test_verifier_ansible_options_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = {}  # type: ignore[var-annotated]

    assert x == _instance.options


@pytest.mark.parametrize(
    "config_instance",
    ["_verifier_section_data"],  # noqa: PT007
    indirect=True,
)
def test_verifier_ansible_options_property_handles_cli_args(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    _instance._config.args = {"debug": True}
    x = {}  # type: ignore[var-annotated]

    assert x == _instance.options


def test_ansible_execute(caplog, _patched_ansible_verify, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    _instance.execute()

    _patched_ansible_verify.assert_called_once_with(None)

    msg = "Running Ansible Verifier"
    assert msg in caplog.text

    msg = "Verifier completed successfully."
    assert msg in caplog.text


def test_execute_does_not_execute(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    patched_ansible_converge,
    caplog: pytest.LogCaptureFixture,
    _instance,  # noqa: PT019
):
    _instance._config.config["verifier"]["enabled"] = False
    _instance.execute()

    assert not patched_ansible_converge.called

    msg = "Skipping, verifier is disabled."
    assert msg in caplog.text



"""Testinfra tests."""  # noqa: INP001

from __future__ import annotations


def test_ansible_hostname(host):  # type: ignore[no-untyped-def]  # noqa: ANN201
    """Validate hostname."""
    f = host.file("/tmp/molecule/instance-1")  # noqa: S108
    assert not f.exists



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#  Copyright (c) 2018 Red Hat, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from molecule import logger
from molecule.app import get_app


if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path


LOG = logger.get_logger(__name__)


@pytest.fixture(name="with_scenario")
def _with_scenario(  # noqa: PLR0913
    request: pytest.FixtureRequest,
    monkeypatch: pytest.MonkeyPatch,
    test_ephemeral_dir_env: dict[str, str],
    scenario_to_test: str,
    scenario_name: str,
    test_fixture_dir: Path,
) -> Iterator[None]:
    """Yield to test with a scenario, changes dir and does cleanup.

    Args:
        request: The pytest request object.
        monkeypatch: The pytest monkeypatch object.
        test_ephemeral_dir_env: The ephemeral directory environment variables.
        scenario_to_test: The scenario to test.
        scenario_name: The scenario name.
        test_fixture_dir: The test fixture directory.

    Yields:
        None:
    """
    scenario_directory = test_fixture_dir / "scenarios" / scenario_to_test

    monkeypatch.chdir(scenario_directory)

    yield
    if request.node.rep_call.failed:
        return
    if scenario_name:
        msg = f"CLEANUP: Destroying instances for {scenario_name}"
        LOG.info(msg)
        cmd = ["molecule", "destroy", "--scenario-name", scenario_name]
        assert (
            get_app(scenario_directory).run_command(cmd=cmd, env=test_ephemeral_dir_env).returncode
            == 0
        )



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import os

import pytest

from molecule import config
from molecule.dependency.ansible_galaxy import roles


@pytest.fixture
def _patched_ansible_galaxy_has_requirements_file(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN202
    m = mocker.patch(
        "molecule.dependency.ansible_galaxy.roles.Roles._has_requirements_file",
    )
    m.return_value = True

    return m


@pytest.fixture
def _dependency_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {
        "dependency": {
            "name": "galaxy",
            "options": {"foo": "bar", "v": True, "requirements-file": "foo.yml"},
            "env": {"FOO": "bar"},
        },
    }


# NOTE(retr0h): The use of the `patched_config_validate` fixture, disables
# config.Config._validate from executing.  Thus preventing odd side-effects
# throughout patched.assert_called unit tests.
@pytest.fixture
def _instance(  # type: ignore[no-untyped-def]  # noqa: ANN202
    _dependency_section_data,
    patched_config_validate,
    config_instance: config.Config,
):
    return roles.Roles(config_instance)


@pytest.fixture(name="role_file")
def fixture_role_file(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return os.path.join(_instance._config.scenario.directory, "requirements.yml")  # noqa: PTH118


def test_roles_config_private_member(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert isinstance(_instance._config, config.Config)


def test_roles_default_options_property(_instance, role_file):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = {"role-file": role_file, "force": False}

    assert x == _instance.default_options


def test_roles_default_env_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    env = _instance.default_env

    assert "MOLECULE_FILE" in env
    assert "MOLECULE_INVENTORY_FILE" in env
    assert "MOLECULE_SCENARIO_DIRECTORY" in env
    assert "MOLECULE_INSTANCE_CONFIG" in env


def test_roles_name_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.name == "galaxy"


def test_roles_enabled_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.enabled


@pytest.mark.parametrize(
    "config_instance",
    ["_dependency_section_data"],  # noqa: PT007
    indirect=True,
)
def test_roles_options_property(_instance, role_file):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = {
        "force": False,
        "role-file": role_file,
        "foo": "bar",
        "v": True,
    }

    assert x == _instance.options


@pytest.mark.parametrize(
    "config_instance",
    ["_dependency_section_data"],  # noqa: PT007
    indirect=True,
)
def test_roles_options_property_handles_cli_args(role_file, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    _instance._config.args = {"debug": True}
    x = {
        "force": False,
        "role-file": role_file,
        "foo": "bar",
        "vvv": True,
    }

    assert x == _instance.options


@pytest.mark.parametrize(
    "config_instance",
    ["_dependency_section_data"],  # noqa: PT007
    indirect=True,
)
def test_roles_env_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.env["FOO"] == "bar"


@pytest.mark.parametrize(
    "config_instance",
    ["_dependency_section_data"],  # noqa: PT007
    indirect=True,
)
def test_galaxy_bake(_instance, role_file):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    _instance.bake()
    args = [
        "ansible-galaxy",
        "install",
        "--foo",
        "bar",
        "--role-file",
        role_file,
        "-v",
    ]
    assert _instance._sh_command == args


def test_execute(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    patched_run_command,
    _patched_ansible_galaxy_has_requirements_file,  # noqa: PT019
    caplog,
    _instance,  # noqa: PT019
):
    _instance._sh_command = "patched-command"
    _instance.execute()

    patched_run_command.assert_called_once_with(
        "patched-command",
        debug=False,
        check=True,
    )

    msg = "Dependency completed successfully."
    assert msg in caplog.text


def test_roles_execute_does_not_execute_when_disabled(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    patched_run_command,
    caplog,
    _instance,  # noqa: PT019
):
    _instance._config.config["dependency"]["enabled"] = False
    _instance.execute()

    assert not patched_run_command.called

    msg = "Skipping, dependency is disabled."
    assert msg in caplog.text


def test_roles_execute_does_not_execute_when_no_requirements_file(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    patched_run_command,
    _patched_ansible_galaxy_has_requirements_file,  # noqa: PT019
    caplog,
    _instance,  # noqa: PT019
):
    _patched_ansible_galaxy_has_requirements_file.return_value = False
    _instance.execute()

    assert not patched_run_command.called

    msg = "Skipping, missing the requirements file."
    assert msg in caplog.text


def test_roles_execute_bakes(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    patched_run_command,
    _instance,  # noqa: PT019
    role_file,
    _patched_ansible_galaxy_has_requirements_file,  # noqa: PT019
):
    _instance.execute()
    assert _instance._sh_command is not None

    assert patched_run_command.call_count == 1


def test_galaxy_executes_catches_and_exits_return_code(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    patched_run_command,
    _patched_ansible_galaxy_has_requirements_file,  # noqa: PT019
    _instance,  # noqa: PT019
):
    patched_run_command.side_effect = SystemExit(1)
    with pytest.raises(SystemExit) as e:
        _instance.execute()

    assert e.value.code == 1


def test_role_file(role_file, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert role_file == _instance.requirements_file


def test_roles_has_requirements_file(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert not _instance._has_requirements_file()



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Base Ansible Galaxy dependency module."""

from __future__ import annotations

import abc
import copy
import os

from pathlib import Path
from typing import TYPE_CHECKING

from molecule import util
from molecule.dependency import base


if TYPE_CHECKING:
    from collections.abc import MutableMapping

    from molecule.config import Config


class AnsibleGalaxyBase(base.Base):
    """Ansible Galaxy dependency base class.

    Attributes:
        FILTER_OPTS: Keys to remove from the dictionary returned by options().
        COMMANDS: Arguments to send to ansible-galaxy to install the appropriate type of content.
    """

    FILTER_OPTS: tuple[str, ...] = ()
    COMMANDS: tuple[str, ...] = ()

    def __init__(self, config: Config) -> None:
        """Construct AnsibleGalaxy.

        Args:
            config: Molecule Config instance.
        """
        super().__init__(config)
        self._sh_command = []

        self.command = "ansible-galaxy"

    @property
    @abc.abstractmethod
    def requirements_file(self) -> str:  # cover
        """Path to requirements file.

        Returns:
            Path to the requirements file for this dependency.
        """

    @property
    def default_options(self) -> MutableMapping[str, str | bool]:
        """Default options for this dependency.

        Returns:
            Default options for this dependency.
        """
        d: MutableMapping[str, str | bool] = {
            "force": False,
        }
        if self._config.debug:
            d["vvv"] = True

        return d

    def filter_options(
        self,
        opts: MutableMapping[str, str | bool],
        keys: tuple[str, ...],
    ) -> MutableMapping[str, str | bool]:
        """Filter certain keys from a dictionary.

        Removes all the values of ``keys`` from the dictionary ``opts``, if
        they are present. Returns the resulting dictionary. Does not modify the
        existing one.

        Args:
            opts: Options dictionary.
            keys: Key names to exclude from opts.

        Returns:
            A copy of ``opts`` without the value of keys
        """
        c = copy.copy(opts)
        for key in keys:
            if key in c:
                del c[key]
        return c

    # NOTE(retr0h): Override the base classes' options() to handle
    # ``ansible-galaxy`` one-off.
    @property
    def options(self) -> MutableMapping[str, str | bool]:
        """Computed options for this dependency.

        Returns:
            Merged and filtered options for this dependency.
        """
        opts = self._config.config["dependency"]["options"]
        # NOTE(retr0h): Remove verbose options added by the user while in
        # debug.
        if self._config.debug:
            opts = util.filter_verbose_permutation(opts)

        opts = util.merge_dicts(self.default_options, opts)
        return self.filter_options(opts, self.FILTER_OPTS)

    @property
    def default_env(self) -> dict[str, str]:
        """Default environment variables for this dependency.

        Returns:
            Default environment variables for this dependency.
        """
        env = dict(os.environ)
        return util.merge_dicts(env, self._config.env)

    def bake(self) -> None:
        """Bake an ``ansible-galaxy`` command so it's ready to execute and returns None."""
        options = self.options
        verbose_flag = util.verbose_flag(options)

        self._sh_command = [
            self.command,
            *self.COMMANDS,
            *util.dict2args(options),
            *verbose_flag,
        ]

    def execute(self, action_args: list[str] | None = None) -> None:  # noqa: ARG002
        """Execute dependency.

        Args:
            action_args: Arguments for this dependency. Unused.
        """
        if not self.enabled:
            msg = "Skipping, dependency is disabled."
            self._log.warning(msg)
            return
        super().execute()

        if not self._has_requirements_file():
            msg = "Skipping, missing the requirements file."
            self._log.warning(msg)
            return

        if not self._sh_command:
            self.bake()

        self._setup()
        self.execute_with_retries()

    def _setup(self) -> None:
        """Prepare the system for using ``ansible-galaxy`` and returns None."""

    def _has_requirements_file(self) -> bool:
        return Path(self.requirements_file).is_file()



"""Test the molecule text module."""

from __future__ import annotations

from molecule.text import (
    camelize,
    strip_ansi_color,
    strip_ansi_escape,
    title,
    underscore,
)


def test_camelize() -> None:  # noqa: D103
    assert camelize("foo") == "Foo"
    assert camelize("foo_bar") == "FooBar"
    assert camelize("foo_bar_baz") == "FooBarBaz"


def test_strip_ansi_color() -> None:  # noqa: D103
    s = "foo\x1b[0m\x1b[0m\x1b[0m\n\x1b[0m\x1b[0m\x1b[0m\x1b[0m\x1b[0m"

    assert strip_ansi_color(s) == "foo\n"


def test_strip_ansi_escape() -> None:  # noqa: D103
    string = "ls\r\n\x1b[00m\x1b[01;31mfoo\x1b[00m\r\n\x1b[01;31m"

    assert strip_ansi_escape(string) == "ls\r\nfoo\r\n"


def test_title() -> None:  # noqa: D103
    assert title("foo") == "Foo"
    assert title("foo_bar") == "Foo Bar"


def test_underscore() -> None:  # noqa: D103
    assert underscore("Foo") == "foo"
    assert underscore("FooBar") == "foo_bar"
    assert underscore("FooBarBaz") == "foo_bar_baz"



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Testinfra Verifier Module."""

from __future__ import annotations

import glob
import os

from pathlib import Path
from typing import TYPE_CHECKING, cast

from molecule import logger, util
from molecule.api import Verifier


if TYPE_CHECKING:
    from collections.abc import MutableMapping

    from molecule.config import Config
    from molecule.verifier.base import Schema


class Testinfra(Verifier):
    """`Testinfra`_ is no longer the default test verifier since version 3.0.

    Additional options can be passed to ``testinfra`` through the options
    dict.  Any option set in this section will override the defaults.

    !!! note

        Molecule will remove any options matching '^[v]+$', and pass ``-vvv``
        to the underlying ``pytest`` command when executing ``molecule
        --debug``.

    ``` yaml
        verifier:
          name: testinfra
          options:
            n: 1
    ```

    The testing can be disabled by setting ``enabled`` to False.

    ``` yaml
        verifier:
          name: testinfra
          enabled: False
    ```

    Environment variables can be passed to the verifier.

    ``` yaml
        verifier:
          name: testinfra
          env:
            FOO: bar
    ```

    Change path to the test directory.

    ``` yaml
        verifier:
          name: testinfra
          directory: /foo/bar/
    ```

    Additional tests from another file or directory relative to the scenario's
    tests directory (supports regexp).

    ``` yaml
        verifier:
          name: testinfra
          additional_files_or_dirs:
            - ../path/to/test_1.py
            - ../path/to/test_2.py
            - ../path/to/directory/*
    ```
    .. _`Testinfra`: https://testinfra.readthedocs.io
    """

    def __init__(self, config: Config) -> None:
        """Set up the requirements to execute ``testinfra`` and returns None.

        Args:
            config: An instance of a Molecule config.
        """
        super().__init__(config)
        self._testinfra_command: list[str] = []
        self._tests = []  # type: ignore[var-annotated]

    @property
    def _log(self) -> logger.ScenarioLoggerAdapter:
        """Get a fresh scenario logger with current context.

        Returns:
            A scenario logger adapter with current scenario and step context.
        """
        # Get step context from the current action being executed
        step_name = getattr(self._config, "action", "verify")
        scenario_name = self._config.scenario.name if self._config else "unknown"
        return logger.get_scenario_logger(__name__, scenario_name, step_name)

    @property
    def name(self) -> str:
        """Name of the verifier.

        Returns:
            The name of the verifier.
        """
        return "testinfra"

    @property
    def default_options(self) -> MutableMapping[str, str | bool]:
        """Get default CLI arguments provided to ``cmd``.

        Returns:
            The default verifier options.
        """
        d = cast("MutableMapping[str, str | bool]", self._config.driver.testinfra_options)
        d["p"] = "no:cacheprovider"
        if self._config.debug:
            d["debug"] = True
            d["vvv"] = True
        if self._config.args.get("sudo"):
            d["sudo"] = True

        return d

    # NOTE(retr0h): Override the base classes' options() to handle
    # ``ansible-galaxy`` one-off.
    @property
    def options(self) -> MutableMapping[str, str | bool]:
        """The computed options for this verifier.

        Returns:
            The combined dictionary of default options and those specified in the config.
        """
        o = self._config.config["verifier"]["options"]
        # NOTE(retr0h): Remove verbose options added by the user while in
        # debug.
        if self._config.debug:
            o = util.filter_verbose_permutation(o)

        return util.merge_dicts(self.default_options, o)

    @property
    def default_env(self) -> dict[str, str]:
        """Get default env variables provided to ``cmd``.

        Returns:
            The default verifier environment variables.
        """
        env = cast("dict[str, str]", os.environ)
        env = util.merge_dicts(env, self._config.env)
        if self._config.provisioner:
            env = util.merge_dicts(env, self._config.provisioner.env)

        return env

    @property
    def additional_files_or_dirs(self) -> list[str]:
        """Additional paths related to the verifier.

        Returns:
            List of files and directories to use with this verifier.
        """
        files_list = []
        c = self._config.config
        for f in c["verifier"]["additional_files_or_dirs"]:
            escaped = glob.escape(self._config.verifier.directory)
            glob_path = os.path.join(escaped, f)  # noqa: PTH118
            glob_list = glob.glob(glob_path)  # noqa: PTH207
            if glob_list:
                files_list.extend(glob_list)

        return files_list

    def bake(self) -> None:
        """Bake a ``testinfra`` command so it's ready to execute."""
        options = self.options
        verbose_flag = util.verbose_flag(options)
        args = verbose_flag

        self._testinfra_command = [
            "pytest",
            *util.dict2args(options),
            *self._tests,
            *args,
        ]

    def execute(self, action_args: list[str] | None = None) -> None:
        """Execute ``cmd``.

        Args:
            action_args: list of arguments to be passed.
        """
        if not self.enabled:
            msg = "Skipping, verifier is disabled."
            self._log.warning(msg)
            return

        if self._config:
            self._tests = self._get_tests(action_args)
        else:
            self._tests = []
        if not len(self._tests) > 0:
            msg = "Skipping, no tests found."
            self._log.warning(msg)
            return

        self.bake()

        msg = f"Executing Testinfra tests found in {self.directory}/..."
        self._log.info(msg)

        result = self._config.app.run_command(
            self._testinfra_command,
            env=self.env,
            debug=self._config.debug,
            cwd=Path(self._config.scenario.directory),
        )
        if result.returncode == 0:
            msg = "Verifier completed successfully."
            self._log.info(msg)
        else:
            util.sysexit(result.returncode)

    def _get_tests(self, action_args: list[str] | None = None) -> list[str]:
        """Walk the verifier's directory for tests.

        Args:
            action_args: List of paths to search.

        Returns:
            List of test files.
        """
        if action_args:
            tests = []
            for arg in action_args:
                args_tests = list(
                    util.os_walk(
                        os.path.join(self._config.scenario.directory, arg),  # noqa: PTH118
                        "test_*.py",
                        followlinks=True,
                    ),
                )
                tests.extend(args_tests)
            return sorted(tests)
        return sorted(
            list(
                util.os_walk(
                    self.directory,
                    "test_*.py",
                    followlinks=True,
                ),
            )
            + self.additional_files_or_dirs,
        )

    def schema(self) -> Schema:
        """Return validation schema.

        Returns:
            Verifier schema.
        """
        return {
            "verifier": {
                "type": "dict",
                "schema": {"name": {"type": "string", "allowed": ["testinfra"]}},
            },
        }



"""Ansible Galaxy dependencies for lists of roles."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, cast

from molecule import util
from molecule.dependency.ansible_galaxy.base import AnsibleGalaxyBase


if TYPE_CHECKING:
    from collections.abc import MutableMapping


class Roles(AnsibleGalaxyBase):
    """Role-specific Ansible Galaxy dependency handling.

    Attributes:
        FILTER_OPTS: Keys to remove from the dictionary returned by options().
        COMMANDS: Arguments to send to ansible-galaxy to install the appropriate type of content.
    """

    FILTER_OPTS = ("requirements-file",)
    COMMANDS = ("install",)

    @property
    def default_options(self) -> MutableMapping[str, str | bool]:
        """Default options for this dependency.

        Returns:
            Default options for this dependency.
        """
        general = super().default_options
        specific = util.merge_dicts(
            general,
            {
                "role-file": str(
                    Path(
                        self._config.scenario.directory,
                        "requirements.yml",
                    ),
                ),
            },
        )
        return specific  # noqa: RET504

    @property
    def requirements_file(self) -> str:
        """Path to requirements file.

        Returns:
            Path to the requirements file for this dependency.
        """
        return cast("str", self.options["role-file"])



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import pytest


@pytest.fixture
def command_patched_ansible_create(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return mocker.patch("molecule.provisioner.ansible.Ansible.create")


@pytest.fixture
def command_driver_delegated_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = {
        "driver": {
            "name": "default",
            "options": {
                "managed": False,
            },
        },
    }
    # if "DOCKER_HOST" in os.environ:
    #     x["driver"]["options"]["ansible_docker_extra_args"] = "-H={}".format(
    return x  # noqa: RET504


@pytest.fixture
def command_driver_delegated_managed_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return {"driver": {"name": "default", "managed": True}}



"""Assorted convenience types for molecule."""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict


if TYPE_CHECKING:
    from collections.abc import MutableMapping
    from typing import Any, Literal, TypeAlias

    Options: TypeAlias = MutableMapping[str, str | bool]


class CollectionData(TypedDict, total=False):
    """Collection metadata sourced from galaxy.yml.

    Attributes:
        name: The name of the collection.
        namespace: The collection namespace.
        version: The collection's version.
        readme: Path to the README file.
        authors: List of authors of the collection.
        description: Description of the collection.
        repository: URL of the collection's online repository.
        license_file: Path to the collection's LICENSE file.
        tags: List of tags applied to the collection.
    """

    name: str
    namespace: str
    version: str
    readme: str
    authors: list[str]
    description: str
    repository: str
    license_file: str
    tags: list[str]


class DependencyData(TypedDict, total=False):
    """Molecule dependency configuration.

    Attributes:
        name: Dependency name.
        command: Command to run for dependency.
        enabled: Is this dependency enabled.
        options: Options for this dependency.
        env: Environment variables related to this dependency.
    """

    name: str
    command: str | None
    enabled: bool
    options: Options
    env: dict[str, str]


class DriverOptions(TypedDict, total=False):
    """Config options for molecule drivers.

    Attributes:
        ansible_connection_options: Options to use with ansible connection plugin.
        login_cmd_template: Template with which to generate login commands.
        managed: Whether the driver is managed.
    """

    ansible_connection_options: dict[str, str]
    login_cmd_template: str
    managed: bool


class DriverData(TypedDict, total=False):
    """Molecule driver configuration.

    Attributes:
        name: Driver name.
        provider:
        options: Options for this driver.
        ssh_connection_options: SSH-specific options.
        safe_files: A list of safe files.
    """

    name: str
    provider: dict[str, Any]
    options: DriverOptions
    ssh_connection_options: list[str]
    safe_files: list[str]


class ExecutorData(TypedDict, total=False):
    """Molecule playbook executor configuration.

    Attributes:
        backend: The backend to use for executing playbooks.
    """

    backend: str


class InventoryData(TypedDict):
    """Inventory data for a molecule run.

    Attributes:
        hosts: A list of hosts.
        host_vars: A dictionary of ansible host_vars.
        group_vars: A dictionary of ansible group_vars.
        links: ???
    """

    hosts: dict[str, str]
    host_vars: dict[str, str]
    group_vars: dict[str, str]
    links: dict[str, str]


class PlatformData(TypedDict, total=False):
    """Platform data for a Molecule run.

    Attributes:
        name: Name of the platform.
        groups: Optional list of groups.
        children: Optional list of child groups.
    """

    name: str
    groups: list[str]
    children: list[str]


class PlaybookData(TypedDict, total=False):
    """Playbooks for a scenario.

    Attributes:
        cleanup: The cleanup playbook.
        create: The create playbook.
        converge: The converge playbook.
        destroy: The destroy playbook.
        prepare: The prepare playbook.
        side_effect: The side_effect playbook.
        verify: The verify playbook.
    """

    cleanup: str
    create: str
    converge: str
    destroy: str
    prepare: str
    side_effect: str
    verify: str


class ProvisionerData(TypedDict, total=False):
    """Molecule provisioner configuration.

    Attributes:
        name: Name of the provisioner.
        config_options: Configuration options.
        ansible_args: Arguments to use with ansible.
        connection_options: Options for the connection.
        options: Options for this provisioner.
        env: Environment variables for this provisioner.
        inventory: Inventory config.
        children: Children of this provisioner.
        playbooks: Playbooks for use in the provisioner.
        log: Should this be logged.
    """

    name: str
    config_options: dict[str, Any]
    ansible_args: list[str]
    connection_options: dict[str, Any]
    options: Options
    env: dict[str, str]
    inventory: InventoryData
    children: dict[str, Any]
    playbooks: PlaybookData
    log: bool


class ScenarioData(TypedDict):
    """Molecule scenario configuration.

    Attributes:
        name: Name of the scenario.
        check_sequence: Sequence of tasks to run for 'check'.
        cleanup_sequence: Sequence of tasks to run for 'cleanup'.
        converge_sequence: Sequence of tasks to run for 'converge'.
        create_sequence: Sequence of tasks to run for 'create'.
        destroy_sequence: Sequence of tasks to run for 'destroy'.
        test_sequence: Sequence of tasks to run for 'test'.
    """

    name: str
    check_sequence: list[str]
    cleanup_sequence: list[str]
    converge_sequence: list[str]
    create_sequence: list[str]
    destroy_sequence: list[str]
    test_sequence: list[str]


class VerifierData(TypedDict, total=False):
    """Molecule verifier configuration.

    Attributes:
        name: Name of the verifier.
        directory: Verifier directory name.
        enabled: Is the verifier enabled.
        options: Options to apply to the verifier.
        env: Applicable environment variables.
        additional_files_or_dirs: Additional paths to verify.
    """

    name: str
    directory: str
    enabled: bool
    options: Options
    env: dict[str, str]
    additional_files_or_dirs: list[str]


class ConfigData(TypedDict):
    """Class representing molecule config.

    Attributes:
        dependency: Dependency config.
        driver: Driver config.
        executor: Executor config.
        platforms: List of platforms.
        prerun: Should prerun tasks be run.
        role_name_check: ???
        provisioner: Provisioner config.
        scenario: Scenario config.
        verifier: Verifier config.
    """

    dependency: DependencyData
    driver: DriverData
    executor: ExecutorData
    platforms: list[PlatformData]
    prerun: bool
    role_name_check: int
    provisioner: ProvisionerData
    scenario: ScenarioData
    verifier: VerifierData


class MoleculeArgs(TypedDict, total=False):
    """Class representing base arguments passed to all Molecule commands.

    Attributes:
        base_config: paths of base config files to load in order.
        debug: Whether to enable debug logging.
        env_file: File to read environment variables from.
        verbose: Verbosity level.
    """

    base_config: list[str]
    debug: bool
    env_file: str
    verbose: int


class CommandArgs(TypedDict, total=False):
    """Class representing commandline arguments passed to molecule.

    These arguments may or may not be passed depending on the command being called.

    Attributes:
        destroy: Destroy strategy to use.
        driver_name: Name of driver to use.
        force: Whether to enable force mode.
        format: Output format
        host: Host to access.
        parallel: Whether to enable parallel mode.
        platform_name: Name of the platform to target.
        report: Whether to show an after-run summary report.
        scenario_name: Name of the scenario to target.
        shared_inventory: Whether inventory should be shared between scenarios.
        shared_state: Whether (some) state should be shared between scenarios.
        subcommand: Name of subcommand being run.
    """

    destroy: Literal["always", "never"]
    driver_name: str
    force: bool
    format: Literal["simple", "plain", "yaml"]
    host: str
    parallel: bool
    platform_name: str
    report: bool
    scenario_name: str
    shared_inventory: bool
    shared_state: bool
    subcommand: str


class ScenarioResult(TypedDict):
    """Dictionary containing the result of a Scenario action.

    Attributes:
        subcommand: The action that ran.
        state: The outcome of the action ("PASSED", "FAILED", or "SKIPPED).
    """

    subcommand: str | None
    state: Literal["PASSED", "FAILED", "SKIPPED"]


class ScenariosResults(TypedDict):
    """Dictionary containing the results of multiple Scenario runs.

    Attributes:
        name: The name of the scenario.
        results: A list of ScenarioResult objects.
    """

    name: str
    results: list[ScenarioResult]



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Matrix Command Module."""

from __future__ import annotations

from typing import TYPE_CHECKING

from molecule import scenarios
from molecule.click_cfg import click_command_ex, options
from molecule.command import base


if TYPE_CHECKING:
    import click

    from molecule.types import CommandArgs


class Matrix(base.Base):
    """Matrix Command Class.

    .. program:: molecule matrix subcommand

    .. option:: molecule matrix subcommand

        Target the default scenario.

    .. program:: molecule matrix --scenario-name foo subcommand

    .. option:: molecule matrix --scenario-name foo subcommand

        Targeting a specific scenario.

    .. program:: molecule --debug matrix subcommand

    .. option:: molecule --debug matrix subcommand

        Executing with `debug`.

    .. program:: molecule --base-config base.yml matrix subcommand

    .. option:: molecule --base-config base.yml matrix subcommand

        Executing with a `base-config`.

    .. program:: molecule --env-file foo.yml matrix subcommand

    .. option:: molecule --env-file foo.yml matrix subcommand

        Load an env file to read variables from when rendering
        molecule.yml.
    """


@click_command_ex()
@options(["scenario_name_single", "subcommand"])
def matrix(ctx: click.Context) -> None:  # pragma: no cover
    """List matrix of steps used to test instances.

    Args:
        ctx: Click context object holding commandline arguments.
    """
    args = ctx.obj.get("args")
    command_args: CommandArgs = {"subcommand": ctx.params["subcommand"]}

    scenario_name = ctx.params["scenario_name"]

    scenario_names = [] if scenario_name is None else [scenario_name]
    s = scenarios.Scenarios(base.get_configs(args, command_args), scenario_names)
    s.print_matrix()



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import copy
import logging
import os

from pathlib import Path
from typing import TYPE_CHECKING, Literal

import pytest

from molecule import config, platforms, scenario, state, util
from molecule.dependency import ansible_galaxy, shell
from molecule.exceptions import MoleculeError
from molecule.provisioner import ansible
from molecule.verifier.ansible import Ansible as AnsibleVerifier


if TYPE_CHECKING:
    from typing import Literal
    from unittest.mock import Mock

    from pytest_mock import MockerFixture

    from molecule.types import DependencyData, DriverData


def test_args_member(config_instance: config.Config) -> None:  # noqa: D103
    assert config_instance.args == {}


def test_command_args_member(config_instance: config.Config) -> None:  # noqa: D103
    x = {"subcommand": "test"}

    assert x == config_instance.command_args


def test_debug_property(config_instance: config.Config) -> None:  # noqa: D103
    assert not config_instance.debug


def test_env_file_property(config_instance: config.Config) -> None:  # noqa: D103
    config_instance.args = {"env_file": ".env"}
    result = config_instance.env_file

    x = config_instance.args.get("env_file")
    assert isinstance(x, str)
    assert util.abs_path(x) == result


def test_subcommand_property(config_instance: config.Config) -> None:  # noqa: D103
    assert config_instance.subcommand == "test"


def test_action_property(config_instance: config.Config) -> None:  # noqa: D103
    assert config_instance.action is None


def test_action_setter(config_instance: config.Config) -> None:  # noqa: D103
    config_instance.action = "foo"

    assert config_instance.action == "foo"


def test_init_calls_validate(  # noqa: D103
    patched_config_validate: Mock,
    config_instance: config.Config,
) -> None:
    patched_config_validate.assert_called_once_with()


def test_collection_directory_property(
    config_instance: config.Config,
    resources_folder_path: Path,
) -> None:
    """Test collection_directory property.

    Args:
        config_instance: Instance of Config.
        resources_folder_path: Path to resources directory holding a valid collection.
    """
    # default path is not in a collection
    assert config_instance.collection_directory is None

    # Alter config_instance to start at path of a collection
    config_instance = copy.copy(config_instance)
    collection_path = resources_folder_path / "sample-collection"
    config_instance.project_directory = str(collection_path)
    assert config_instance.collection_directory == collection_path


def test_project_directory_property(config_instance: config.Config) -> None:  # noqa: D103
    assert str(Path.cwd()) == config_instance.project_directory


def test_molecule_directory_property(config_instance: config.Config) -> None:  # noqa: D103
    x = str(Path.cwd() / "molecule")

    assert x == config_instance.molecule_directory


def test_collection_property(
    config_instance: config.Config,
    resources_folder_path: Path,
) -> None:
    """Test collection property.

    Args:
        config_instance: Instance of Config.
        resources_folder_path: Path to resources directory holding a valid collection.
    """
    modified_instance = copy.copy(config_instance)
    # default path is not in a collection
    assert config_instance.collection is None

    # Alter config_instance to start at path of a collection
    collection_path = resources_folder_path / "sample-collection"
    modified_instance.project_directory = str(collection_path)

    assert modified_instance.collection is not None
    assert modified_instance.collection["name"] == "goodies"
    assert modified_instance.collection["namespace"] == "acme"


def test_collection_property_broken_collection(
    caplog: pytest.LogCaptureFixture,
    config_instance: config.Config,
    resources_folder_path: Path,
) -> None:
    """Test collection property with a malformed galaxy.yml.

    Args:
        caplog: pytest log capture fixture.
        config_instance: Instance of Config.
        resources_folder_path: Path to resources directory holding a valid collection.
    """
    modified_instance = copy.copy(config_instance)

    # Alter config_instance to start at path of a collection
    collection_path = resources_folder_path / "broken-collection"
    modified_instance.project_directory = str(collection_path)

    assert modified_instance.collection is None

    msg = "missing mandatory field 'namespace'"
    assert msg in caplog.text


def test_dependency_property(config_instance: config.Config) -> None:  # noqa: D103
    assert isinstance(config_instance.dependency, ansible_galaxy.AnsibleGalaxy)


@pytest.fixture
def _config_dependency_shell_section_data() -> dict[Literal["dependency"], DependencyData]:
    return {"dependency": {"name": "shell", "command": "bin/command"}}


@pytest.mark.parametrize(
    "config_instance",
    ["_config_dependency_shell_section_data"],  # noqa: PT007
    indirect=True,
)
def test_dependency_property_is_shell(config_instance: config.Config) -> None:  # noqa: D103
    assert isinstance(config_instance.dependency, shell.Shell)


@pytest.fixture
def _config_driver_delegated_section_data() -> dict[Literal["driver"], DriverData]:
    return {"driver": {"name": "default", "options": {"managed": False}}}


def test_env(config_instance: config.Config) -> None:  # noqa: D103
    config_instance.args = {"env_file": ".env"}
    env_file = config_instance.args.get("env_file")
    assert isinstance(env_file, str)
    x = {
        "MOLECULE_DEBUG": "False",
        "MOLECULE_FILE": config_instance.config_file,
        "MOLECULE_ENV_FILE": util.abs_path(env_file),
        "MOLECULE_INVENTORY_FILE": config_instance.provisioner.inventory_file,  # type: ignore[union-attr]
        "MOLECULE_EPHEMERAL_DIRECTORY": config_instance.scenario.ephemeral_directory,
        "MOLECULE_SCENARIO_DIRECTORY": config_instance.scenario.directory,
        "MOLECULE_PROJECT_DIRECTORY": config_instance.project_directory,
        "MOLECULE_INSTANCE_CONFIG": config_instance.driver.instance_config,
        "MOLECULE_DEPENDENCY_NAME": "galaxy",
        "MOLECULE_DRIVER_NAME": "default",
        "MOLECULE_PROVISIONER_NAME": "ansible",
        "MOLECULE_SCENARIO_NAME": "default",
        "MOLECULE_STATE_FILE": config_instance.state.state_file,
        "MOLECULE_VERIFIER_NAME": "ansible",
        "MOLECULE_VERIFIER_TEST_DIRECTORY": config_instance.verifier.directory,
    }

    assert x == config_instance.env


def test_platforms_property(config_instance: config.Config) -> None:  # noqa: D103
    assert isinstance(config_instance.platforms, platforms.Platforms)


def test_provisioner_property(config_instance: config.Config) -> None:  # noqa: D103
    assert isinstance(config_instance.provisioner, ansible.Ansible)


def test_scenario_property(config_instance: config.Config) -> None:  # noqa: D103
    assert isinstance(config_instance.scenario, scenario.Scenario)


def test_state_property(config_instance: config.Config) -> None:  # noqa: D103
    assert isinstance(config_instance.state, state.State)


def test_verifier_property_is_ansible(config_instance: config.Config) -> None:  # noqa: D103
    assert isinstance(config_instance.verifier, AnsibleVerifier)


def test_verifier_property_invalid(config_instance: config.Config) -> None:  # noqa: D103
    config_instance.config["verifier"]["name"] = "missing"
    del config_instance.verifier

    with pytest.raises(RuntimeError, match="Unable to find 'missing' verifier driver."):
        config_instance.verifier  # noqa: B018


def test_get_driver_name_from_state_file(  # noqa: D103
    config_instance: config.Config,
    mocker: MockerFixture,
) -> None:
    config_instance.state.change_state("driver", "state-driver")

    with pytest.raises(MoleculeError):
        config_instance._get_driver_name()

    mocker.patch("molecule.api.drivers", return_value=["state-driver"])
    assert config_instance._get_driver_name() == "state-driver"


def test_get_driver_name_from_cli(config_instance: config.Config) -> None:  # noqa: D103
    config_instance.command_args = {"driver_name": "cli-driver"}

    assert config_instance._get_driver_name() == "cli-driver"


def test_get_driver_name(config_instance: config.Config) -> None:  # noqa: D103
    assert config_instance._get_driver_name() == "default"


def test_get_driver_name_raises_when_different_driver_used(  # noqa: D103
    caplog: pytest.LogCaptureFixture,
    config_instance: config.Config,
) -> None:
    config_instance.state.change_state("driver", "foo")
    config_instance.command_args = {"driver_name": "bar"}
    with pytest.raises(MoleculeError) as e:
        config_instance._get_driver_name()

    assert e.value.code == 1

    msg = (
        "Instance(s) were created with the 'foo' driver, but the subcommand is using 'bar' driver."
    )

    assert msg in caplog.text


def test_get_config(config_instance: config.Config) -> None:  # noqa: D103
    assert isinstance(config_instance._get_config(), dict)


def test_get_config_with_base_config(config_instance: config.Config) -> None:  # noqa: D103
    config_instance.args = {"base_config": ["./foo.yml"]}
    contents = {"foo": "bar"}
    util.write_file(config_instance.args["base_config"][0], util.safe_dump(contents))
    result = config_instance._get_config()

    assert result["foo"] == "bar"  # type: ignore[typeddict-item]


def test_get_config_with_multiple_base_configs(  # noqa: D103
    config_instance: config.Config,
) -> None:
    config_instance.args = {"base_config": ["./foo.yml", "./foo2.yml"]}
    contents = {"foo": "bar", "foo2": "bar"}
    util.write_file(config_instance.args["base_config"][0], util.safe_dump(contents))
    contents = {"foo2": "bar2"}
    util.write_file(config_instance.args["base_config"][1], util.safe_dump(contents))
    result = config_instance._get_config()

    assert result["foo"] == "bar"  # type: ignore[typeddict-item]
    assert result["foo2"] == "bar2"  # type: ignore[typeddict-item]


def test_reget_config(config_instance: config.Config) -> None:  # noqa: D103
    assert isinstance(config_instance._reget_config(), dict)


def test_interpolate(config_instance: config.Config) -> None:  # noqa: D103
    string = "foo: $HOME"
    x = f"foo: {os.environ['HOME']}"

    assert x == config_instance._interpolate(string, os.environ, "")


def test_interpolate_curly(config_instance: config.Config) -> None:  # noqa: D103
    string = "foo: ${HOME}"
    x = f"foo: {os.environ['HOME']}"

    assert x == config_instance._interpolate(string, os.environ, "")


def test_interpolate_default(config_instance: config.Config) -> None:  # noqa: D103
    string = "foo: ${NONE-default}"
    x = "foo: default"

    assert x == config_instance._interpolate(string, os.environ, "")


def test_interpolate_default_colon(config_instance: config.Config) -> None:  # noqa: D103
    string = "foo: ${NONE:-default}"
    x = "foo: default"

    assert x == config_instance._interpolate(string, os.environ, "")


def test_interpolate_default_variable(config_instance: config.Config) -> None:  # noqa: D103
    string = "foo: ${NONE:-$HOME}"
    x = f"foo: {os.environ['HOME']}"

    assert x == config_instance._interpolate(string, os.environ, "")


def test_interpolate_curly_default_variable(  # noqa: D103
    config_instance: config.Config,
) -> None:
    string = "foo: ${NONE-$HOME}"
    x = f"foo: {os.environ['HOME']}"

    assert x == config_instance._interpolate(string, os.environ, "")


def test_interpolate_raises_on_failed_interpolation(  # noqa: D103
    caplog: pytest.LogCaptureFixture,
    config_instance: config.Config,
) -> None:
    string = "$6$8I5Cfmpr$kGZB"

    with pytest.raises(MoleculeError) as e:
        config_instance._interpolate(string, os.environ, "")

    assert e.value.code == 1

    msg = (
        f"parsing config file '{config_instance.molecule_file}'.\n\n"
        "Invalid placeholder in string: line 1, col 1\n"
        "$6$8I5Cfmpr$kGZB"
    )
    assert msg in caplog.text


def test_get_defaults(  # noqa: D103
    config_instance: config.Config,
    mocker: MockerFixture,
) -> None:
    mocker.patch.object(
        config_instance,
        "molecule_file",
        "/path/to/test_scenario_name/molecule.yml",
    )
    defaults = config_instance._get_defaults()
    assert defaults["scenario"]["name"] == "test_scenario_name"


def test_validate(  # noqa: D103
    mocker: MockerFixture,
    config_instance: config.Config,
    caplog: pytest.LogCaptureFixture,
) -> None:
    m = mocker.patch("molecule.model.schema_v3.validate")
    m.return_value = None

    with caplog.at_level(logging.DEBUG):
        config_instance._validate()

    # Check that debug message was logged
    assert len(caplog.records) == 1

    # Check the log message content
    record = caplog.records[0]
    assert "Validating schema" in record.message
    assert config_instance.molecule_file in record.message

    # Check that scenario logger extras are present
    assert hasattr(record, "molecule_scenario")
    assert hasattr(record, "molecule_step")
    assert record.molecule_step == "validate"

    m.assert_called_with(config_instance.config)


def test_validate_exists_when_validation_fails(  # noqa: D103
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
    config_instance: config.Config,
) -> None:
    m = mocker.patch("molecule.model.schema_v3.validate")
    m.return_value = "validation errors"

    with pytest.raises(MoleculeError) as e:
        config_instance._validate()

    assert e.value.code == 1

    msg = f"Failed to validate {config_instance.molecule_file}\n\nvalidation errors"
    assert msg in caplog.text


def test_molecule_directory() -> None:  # noqa: D103
    assert config.molecule_directory("/foo/bar") == "/foo/bar/molecule"


def test_molecule_file() -> None:  # noqa: D103
    assert config.molecule_file("/foo/bar") == "/foo/bar/molecule.yml"


def test_set_env_from_file(config_instance: config.Config) -> None:  # noqa: D103
    config_instance.args = {"env_file": ".env"}
    contents = {"foo": "bar", "BAZ": "zzyzx"}
    env_file = config_instance.args.get("env_file")
    assert isinstance(env_file, str)
    util.write_file(env_file, util.safe_dump(contents))
    env = config.set_env_from_file({}, env_file)

    assert contents == env


def test_set_env_from_file_returns_original_env_when_env_file_not_found(  # noqa: D103
    config_instance: config.Config,
) -> None:
    env = config.set_env_from_file({}, "file-not-found")

    assert env == {}


def test_write_config(config_instance: config.Config) -> None:  # noqa: D103
    config_instance.write()

    assert os.path.isfile(config_instance.config_file)  # noqa: PTH113
